VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsBitmap"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' -------------------------------------------------------------------------------
'
'  clsLog - A class for creating application log files
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   clsLog.cls
'   Purpose: In-memory bitmap functions.
'
Option Explicit
Option Compare Binary
Option Base 0


' VB Array desriptor
Private Type SafeArrayBound
    cElements As Long
    lLbound As Long
End Type

Private Type SafeArray2D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    rgSABounds(0 To 1) As SafeArrayBound
End Type

' RGB bitmap palette
Private Type RGBQUAD
    rgbBlue As Byte
    rgbGreen As Byte
    rgbRed As Byte
    rgbReserved As Byte
End Type

' Bitmap Info Header
Private Type BitmapInfoHeader
     biSize As Long
     biWidth As Long
     biHeight As Long
     biPlanes As Integer
     biBitCount As Integer
     biCompression As Long
     biSizeImage As Long
     biXPelsPerMeter As Long
     biYPelsPerMeter As Long
     biClrUsed As Long
     biClrImportant As Long
End Type

Private Type BitmapInfo
    biHeader As BitmapInfoHeader
    biColors(0 To 255) As RGBQUAD
End Type


' Bitmap header data
Private DIBInfo As BitmapInfo
Private Palette(0 To 255) As RGBQUAD

' Bitmap data
Private DataPointer As Long
Private Stride As Long

' Handles to bitmap and device contexts
Private ihBitmap As Long
Private ihDC As Long


' API calls
Private Declare Sub PutDWord Lib "MSVBVM60.dll" Alias "PutMem4" (ByRef inDst As Any, ByVal inSrc As Long)
Private Declare Function CreateDIBSection Lib "GDI32.dll" (ByVal hDC As Long, ByRef pBMI As BitmapInfo, ByVal iUsage As Long, ByRef ppvBits As Long, ByVal hSection As Long, ByVal dwOffset As Long) As Long
Private Declare Function CreateCompatibleDC Lib "gdi32" (ByVal hDC As Long) As Long
Private Declare Function VarPtrArray Lib "MSVBVM60.dll" Alias "VarPtr" (ByRef Ptr() As Any) As Long
Private Declare Function DeleteObject Lib "GDI32.dll" (ByVal hObject As Long) As Long
Private Declare Function DeleteDC Lib "GDI32.dll" (ByVal hDC As Long) As Long
Private Declare Function SelectObject Lib "GDI32.dll" (ByVal hDC As Long, ByVal hObject As Long) As Long



' Device Context property
Public Property Get hDC() As Long

    hDC = ihDC

End Property


' Bitmap handle property
Public Property Get hBitmap() As Long

    hBitmap = ihBitmap

End Property


' Inject RAW data straight into the bitmap
Public Sub Inject(ByRef Data As String)

    CopyMemory ByVal DataPointer, ByVal Data, Len(Data)

End Sub


' Set the palette of the bitmap
Public Sub setPalette(ByRef Data As String)

    Dim A As Long
    
    
    ' Create the bitmap palette
    For A = 0 To 255
        Palette(A).rgbRed = Asc(Mid(Data, (A * 3) + 1, 1))
        Palette(A).rgbGreen = Asc(Mid(Data, (A * 3) + 2, 1))
        Palette(A).rgbBlue = Asc(Mid(Data, (A * 3) + 3, 1))
    Next A

End Sub


' Create an empty paletted bitmap
Public Sub Create(ByVal Width As Long, ByVal Height As Long, ByVal fillColor As Byte)
    
    Dim A As Long
    
    
    ' Set DIB header info
    With DIBInfo.biHeader
        .biSize = Len(DIBInfo.biHeader)
        .biWidth = Width
        .biHeight = Height
        .biBitCount = 8
        .biPlanes = 1
        .biSizeImage = alignScan(.biWidth, .biBitCount) * .biHeight
        .biClrUsed = 256
        .biClrImportant = 256
        
        Stride = alignScan(.biWidth, .biBitCount)
    End With
    
    ' Set palette
    For A = 0 To 255
        DIBInfo.biColors(A) = Palette(A)
    Next A

    ' Create the DIB
    ihDC = CreateCompatibleDC(0)
    ihBitmap = CreateDIBSection(ihDC, DIBInfo, 0, DataPointer, 0, 0)
    SelectObject ihDC, ihBitmap

   ' Fill with transparency color
    FillMemory ByVal DataPointer, DIBInfo.biHeader.biSizeImage, fillColor

End Sub


' Clean up
Public Sub Destroy()

    ' Deselect and delete
    SelectObject ihDC, 0
    
    DeleteDC ihDC
    DeleteObject ihBitmap
    
End Sub


' Render a Doom image onto this bitmap
' I forgot who wrote the code originally, I only modified it for my needs
Public Sub renderDoomImage(ByRef Data As String, ByVal drawX As Long, ByVal drawY As Long)

    Dim dataBytes() As Byte
    Dim X As Long
    Dim Y As Long
    
    Dim Columns As Integer
    Dim Rows As Integer
    Dim columnPtr() As Long
    
    Dim I As Long, C As Long, rS As Long, rP As Long, Pxls As Long


    ' Store the image data also as byte array
    ReDim dataBytes(1 To Len(Data)) As Byte
    CopyMemory dataBytes(1), ByVal Data, Len(Data)
    
    ' Read the header from Data
    Columns = CVI(Mid(Data, 1, 2))
    Rows = CVI(Mid(Data, 3, 2))


    ' Read column pointers from Data
    ReDim columnPtr(0 To Columns - 1) As Long
    For I = 0 To Columns - 1
        columnPtr(I) = CVL(Mid(Data, 5 + (I + 1) * 4, 4))
    Next I
    
    ' Create the actual bitmap data
    For C = 0 To Columns - 1
        
        ' Start at position given by column pointer
        I = columnPtr(C) + 1
        
        ' Read the start row
        rS = dataBytes(I)
        
        ' Continue until column is finished
        Do Until rS = 255
        
            ' Read the number of pixels to draw
            Pxls = Asc(Mid(Data, I + 1, 1))
            
            ' Go to start position of the pixels
            I = I + 2
            
            ' Read each pixel and 'draw' it
            X = drawX + C
            If X >= 0 And X < DIBInfo.biHeader.biWidth Then

                For rP = 1 To Pxls
                    Y = drawY + (rS + rP) - 1
                    If Y >= 0 And Y < DIBInfo.biHeader.biHeight Then
                        CopyMemory ByVal DataPointer + X + ((DIBInfo.biHeader.biHeight - Y - 1) * Stride), ByVal VarPtr(dataBytes(I + rP)), 1
                    End If
                Next rP

            End If
            
            ' Shift i to the next counter's position
            I = I + Pxls + 2
            
            ' Read next start row
            rS = dataBytes(I)
            
        Loop
        
    Next C
        
End Sub


' DWord-alignment function
Private Function alignScan(ByVal inWidth As Long, ByVal inDepth As Integer) As Long
    
    alignScan = (((inWidth * inDepth) + &H1F) And Not &H1F&) \ &H8
    
End Function


' Class
Private Sub Class_Terminate()

    Me.Destroy

End Sub
