VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsLog"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' -------------------------------------------------------------------------------
'
'  clsLog - A class for creating application log files
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   clsLog.cls
'   Purpose: Application logging class.
'
'   Usage:
'
'     - Create a new log file using StartLog (Append appends log to existing if any).
'     - Write log lines using Add method.
'     - Close log during runtime using CloseLog if necessary, but log is closed on
'       class termination also.
'
Option Explicit
Option Compare Binary
Option Base 0


' Log file info
Private LogFile As String
Private LogFileNum As Long


' Start logging
Public Sub StartLog(ByVal NewLogFile As String, ByVal Append As Boolean)

    CloseLog
    
    ' Remove old log if not appending
    If Append = False And FS.FileExists(NewLogFile) Then FS.DeleteFile NewLogFile
    
    ' Reopen the new one
    LogFile = NewLogFile
    LogFileNum = FreeFile
    Open NewLogFile For Append As LogFileNum
    
    Print #LogFileNum, ""
    Print #LogFileNum, String$(70, "-")
    Print #LogFileNum, "Append log at " & Now
    Me.Add App.Title & " version " & App.Major & "." & App.Minor & " build " & App.Revision
    Me.Add String$(58, "-")

End Sub


' Add something to the log
Public Sub Add(ByVal WriteString As String)

    ' Exit if no log file opened
    If LogFileNum = -1 Then Exit Sub
    
    Print #LogFileNum, "[" & Time$ & "]  " & WriteString

End Sub


' Class intialisation
Private Sub Class_Initialize()

    LogFileNum = -1

End Sub


' Close any open log file
Private Sub Class_Terminate()

    CloseLog
    
End Sub


' Close any possible open log file
Private Sub CloseLog()

    If LogFileNum <> -1 Then
        Close LogFileNum
        LogFileNum = -1
    End If

End Sub
