VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsWad"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' -------------------------------------------------------------------------------
'
'  clsWad - A simple Doom WAD file handling class
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   clsWad.cls
'   Purpose: Wad handling class, contains most of the wad manipulating
'            routines a wad editing program can use.
'
'   Notes: Some parts are written especially for WhackEd3.
'
'   TODO: Nothing.
'
Option Explicit
Option Compare Binary
Option Base 0


' A lump
Private Type Lump
    offset As Long
    Size As Long
    name As String * 8
End Type

' Wad data
Private iLumps() As Lump
Private iWadType As String * 4
Private iWadLumps As Long
Private iWadDirOffset As Long
Private iWadFileName As String
Private iWadFileNumber As Integer

' Sprite lumps
Private SpriteList() As Long
Private nSprites As Long

' Wad IDs
Private Const WAD_PWAD As String * 4 = "PWAD"
Private Const WAD_IWAD As String * 4 = "IWAD"


' --------------------------------------------------------------------------------
'  SUB          : OpenWad
'  DESCRIPTION  : Open a new wad file and load its lumps
' --------------------------------------------------------------------------------
Public Sub OpenWad(ByVal filename As String)

    Dim openFile As Long

    Dim cWadId As String * 4
    Dim cWadDirOffset As Long
    Dim cWadLumps As Long


    ' Do some validations
    If Not FS.FileExists(filename) Then Exit Sub
    If iWadFileNumber > -1 Then EraseData

    ' Open the wad
    openFile = FreeFile
    Open filename For Binary As openFile

        ' Read the header
        Get #openFile, , cWadId
        Get #openFile, , cWadLumps
        Get #openFile, , cWadDirOffset

        ' See if the header is valid
        If (cWadId <> WAD_IWAD) And (cWadId <> WAD_PWAD) Then
            Close openFile
            Exit Sub
        End If

        ' Erase the old class wad data
        EraseData

        ' Load the lump list if possible
        If cWadLumps <> -1 Then
            ReDim iLumps(cWadLumps)
            Seek #openFile, cWadDirOffset + 1
            Get #openFile, , iLumps()
        End If

        ' Set the wad data
        iWadType = cWadId
        iWadLumps = cWadLumps - 1
        iWadDirOffset = cWadDirOffset
        iWadFileName = filename
        iWadFileNumber = openFile
        
        ' Create the sprite list index
        CreateSpriteIndex

End Sub


' --------------------------------------------------------------------------------
'  FUNCTION     : ReadLump
'  DESCRIPTION  : Read a lump from a wad and return the data
' --------------------------------------------------------------------------------
Public Function ReadLump(ByVal LumpNum As Long) As Byte()

    If LumpNum = -1 Then Exit Function

    ' Read the lump
    ReDim ReadLump(iLumps(LumpNum).Size)
    Seek iWadFileNumber, iLumps(LumpNum).offset + 1
    Get iWadFileNumber, , ReadLump

End Function


' --------------------------------------------------------------------------------
'  FUNCTION     : ReadLumpString
'  DESCRIPTION  : Read a lump from a wad and return the data (in String format)
' --------------------------------------------------------------------------------
Public Function ReadLumpString(ByVal LumpNum As Long) As String

    If LumpNum = -1 Then Exit Function

    ' Read the lump
    ReadLumpString = Space$(iLumps(LumpNum).Size)
    Seek iWadFileNumber, iLumps(LumpNum).offset + 1
    Get iWadFileNumber, , ReadLumpString

End Function


' --------------------------------------------------------------------------------
'  FUNCTION     : SeekLump
'  DESCRIPTION  : Seek to the offset of a lump and return the wad's filenumber
' --------------------------------------------------------------------------------
Public Function SeekLump(ByVal LumpIndex As Long) As Long

    ' Seek and return the filenumber
    Seek iWadFileNumber, iLumps(LumpIndex).offset + 1
    SeekLump = iWadFileNumber

End Function


' --------------------------------------------------------------------------------
'  SUB          : EraseData (PRIVATE)
'  DESCRIPTION  : Clear all current wad data
' --------------------------------------------------------------------------------
Private Sub EraseData()

    If iWadFileNumber <> -1 Then Close iWadFileNumber
    iWadType = WAD_PWAD
    iWadLumps = -1
    iWadDirOffset = 12
    iWadFileName = vbNullString
    iWadFileNumber = -1
    ReDim iLumps(0)

End Sub


' --------------------------------------------------------------------------------
'  PROPERTY     : WadType
'  METHODS      : Get\Let
'  DESCRIPTION  : The type of wad (IWAD or PWAD)
' --------------------------------------------------------------------------------
Public Property Get WadType() As Long

    ' Return a proper wad type value
    Select Case iWadType
        Case "IWAD"
            WadType = 0
        Case "PWAD"
            WadType = 1
        Case Else
            WadType = -1
    End Select

End Property

Public Property Let WadType(ByVal NewType As Long)

    ' Set the proper wad type
    Select Case NewType
        Case 0
            iWadType = WAD_IWAD
        Case 1
            iWadType = WAD_PWAD
    End Select

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : WadFileName
'  METHODS      : Get
'  DESCRIPTION  : The filename of the wad we're using
' --------------------------------------------------------------------------------
Public Property Get WadFileName() As String

    WadFileName = iWadFileName

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : WadLumps
'  METHODS      : Get
'  DESCRIPTION  : The number of lumps the currently loaded wad has
' --------------------------------------------------------------------------------
Public Property Get WadLumps() As String

    WadLumps = iWadLumps

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : LumpName
'  METHODS      : Get
'  DESCRIPTION  : The name of a lump
' --------------------------------------------------------------------------------
Public Property Get LumpName(ByVal nLump As Long) As String

    LumpName = StripNull(iLumps(nLump).name)

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : LumpSize
'  METHODS      : Get
'  DESCRIPTION  : The size of a lump
' --------------------------------------------------------------------------------
Public Property Get LumpSize(ByVal nLump As Long) As Long

    LumpSize = iLumps(nLump).Size

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : LumpOffset
'  METHODS      : Get
'  DESCRIPTION  : The offset of a lump
' --------------------------------------------------------------------------------
Public Property Get LumpOffset(ByVal nLump As Long) As Long

    LumpOffset = iLumps(nLump).offset

End Property


' --------------------------------------------------------------------------------
'  PROPERTY     : WadSize
'  METHODS      : Get
'  DESCRIPTION  : The size of the wad file
' --------------------------------------------------------------------------------
Public Property Get WadSize() As Long

    WadSize = LOF(iWadFileNumber)

End Property


' --------------------------------------------------------------------------------
'  FUNCTION     : NumForName
'  DESCRIPTION  : Get a lump index number from a lump name
' --------------------------------------------------------------------------------
Public Function NumForName(ByVal LumpName As String) As Long

    Dim A As Long


    ' Search for the lump
    LumpName = UCase(Trim$(LumpName))
    For A = 0 To iWadLumps
        If LumpName = StripNull(iLumps(A).name) Then
            NumForName = A
            Exit Function
        End If
    Next A

    ' We didn't find the lump
    NumForName = -1

End Function


' --------------------------------------------------------------------------------
'  FUNCTION     : NumForSprite
'  DESCRIPTION  : Get a lump index number from a lump name
' --------------------------------------------------------------------------------
Public Function NumForSprite(ByVal SpriteName As String, ByVal FrameRot As String) As Long

    Dim A As Long
    Dim lName As String

    
    ' Search for the sprite
    For A = 0 To nSprites
        
        lName = iLumps(SpriteList(A)).name
        
        ' See if this is the right sprite
        If Left$(lName, 4) = SpriteName Then
            
            ' Normal
            If Mid$(lName, 5, 2) = FrameRot Then
                NumForSprite = SpriteList(A)
                Exit Function
            
            ' Mirrored
            ElseIf Mid$(lName, 7, 2) = FrameRot Then
                NumForSprite = SpriteList(A)
                Exit Function
                
            End If
                
        End If
        
    Next A

    ' We didn't find the lump
    NumForSprite = -1

End Function


' --------------------------------------------------------------------------------
'  FUNCTION     : CreateSpriteIndex
'  DESCRIPTION  : Create a list of sprites in the WAD
' --------------------------------------------------------------------------------
Public Sub CreateSpriteIndex()

    Dim A As Long
    Dim InList As Long
    Dim LumpName As String
    
    
    ' Reset sprite list
    ReDim SpriteList(0)
    nSprites = -1
    
    
    For A = 0 To iWadLumps
        
        ' Get the name of the lump we're examining
        LumpName = StripNull(iLumps(A).name)

        ' Found the start of the sprites list
        If LumpName = "S_START" Or LumpName = "SS_START" Then
            InList = 1
            GoTo SkipLump
        End If

        ' Found the end of the sprite list
        If LumpName = "S_END" Or LumpName = "SS_END" Then
            InList = 0
            Exit Sub
        End If

        ' If we are inside the sprite list, add the sprite to the list
        If InList = 1 Then
            nSprites = nSprites + 1
            ReDim Preserve SpriteList(nSprites)
            SpriteList(nSprites) = A
        End If
        
        
SkipLump:
        
    Next A

End Sub


' --------------------------------------------------------------------------------
'  FUNCTION     : StripNull (PRIVATE)
'  DESCRIPTION  : Strip a string from NULL characters, wrapper
' --------------------------------------------------------------------------------
Private Function StripNull(ByVal InString As String) As String

    StripNull = Trim$(Replace(InString, Chr$(0), " "))

End Function


' --------------------------------------------------------------------------------
'  FUNCTION     : PadNull (PRIVATE)
'  DESCRIPTION  : Pad a string with NULL characters, semi-wrapper
' --------------------------------------------------------------------------------
Private Function PadNull(ByVal InString As String, PadCount As Long) As String

    InString = Trim$(InString)
    PadNull = InString & String(PadCount - Len(InString), Chr$(0))

End Function


' --------------------------------------------------------------------------------
'  SUB          : Class
'  DESCRIPTION  : -
' --------------------------------------------------------------------------------
Private Sub Class_Initialize()

    EraseData

End Sub

Private Sub Class_Terminate()

    EraseData

End Sub
