Attribute VB_Name = "modConfig"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modConfig.bas
'   Purpose: Configuration handling module.
'
Option Explicit
Option Compare Binary
Option Base 0


' Config structs
Public Config As New clsConfiguration

' Internal variables
Public VersionString As String
Public ConfigFile As String

' Dictionary of engine configuration files
Public ConfigFiles As New Dictionary
Public ConfigFilesData As New Collection

' Configuration type identifiers
Public Const CFGTYPE_ENGINE As String = "WhackEd3 Engine Configuration file"
Public Const CFGTYPE_PROGRAM As String = "WhackEd3 Program Configuration file"
Public Const CFGTYPE_IWADS As String = "WhackEd3 IWAD detect configuration file"


' Set program defaults
Public Sub SetConfigDefaults()

    ' Internal variables
    VersionString = "WhackEd3 " & App.Major & "." & App.Minor & " build " & App.Revision
    If BETAMESSAGE = True Then VersionString = VersionString & " beta"
    ConfigFile = DATA_PATH & "whacked3.cfg"

End Sub


' Load program configuration
Public Sub LoadProgramConfig()

    Dim A As Long
    Dim Removed As Boolean


    ' See if the configuration file exists
    If Not FS.FileExists(ConfigFile) Then modConfig.WriteNewConfig

    ' Read the config
    Config.LoadConfiguration ConfigFile

    ' Validate it
    If Config.Root(False)("Type") <> CFGTYPE_PROGRAM Then
        MsgBox "Invalid program configuration file.", vbCritical Or vbOKOnly
        modMain.Quit
    End If


    ' Check if the user is running the program for the first time
    If Config.Root(False)("firstRun") = 1 Then

        ' Display first run info
        MsgBox "This is the first time WhackEd3 is run from this location. Please adjust the program's settings and be sure to add at least one IWAD.", vbOKOnly Or vbInformation

        appLog.Add "First run, displaying Options dialog."

        frmOptions.initConfig
        frmOptions.Show 1

        Config.Root(True)("firstRun") = 0
        modConfig.SaveProgramConfig

        frmSplash.Refresh

    End If
    
    ' Filter out nonexistent recent files
    Do
        Removed = False
        For A = 0 To RecentFiles.Count - 1
            If Not FS.FileExists(RecentFiles.Items(A)) Then
                RecentFiles.Remove Trim$(Str$(RecentFiles.Keys(A)))
                Removed = True
                Exit For
            End If
        Next A
    Loop Until Removed = False

End Sub


' Save the program configuration
Public Sub SaveProgramConfig()

    ' Save the config
    Config.SaveConfiguration ConfigFile, , True

End Sub


' Write a new configuration file with defaults
Private Sub WriteNewConfig()

    With Config.Root(True)
    
        ' Type
        .Add "Type", CFGTYPE_PROGRAM
        
        ' Subclasses
        .Add "IWADs", New Dictionary
        .Add "recentFiles", New Dictionary

        .Add "createBackup", 1
        .Add "stateActivate", 0
        .Add "frameCharacter", 0
        
        .Add "firstRun", 1
        .Add "lastDir", getSpecialPath(myDocuments)
        .Add "lastConfig", 0
        .Add "lastIWAD", 0
        
        ' Recent files
        ' (empty by default)
    
    End With

    ' Write and save them
    Config.SaveConfiguration ConfigFile

End Sub


' Load an engine configuration file
Public Function LoadEngineConfig(ByVal filename As String) As Boolean

    Dim A As Long
    
    Dim EngineConfig As New clsConfiguration
    Dim ConfigRoot As Dictionary
    Dim TempDic As Dictionary
    
    
    ' Load the configuration
    EngineConfig.LoadConfiguration filename
    
    ' Validate it
    If EngineConfig.Root.Item("type") <> CFGTYPE_ENGINE Then
        MsgBox "Invalid engine configuration file.", vbExclamation Or vbOKOnly
        LoadEngineConfig = False
        Exit Function
    End If
    
    
    ' Read all settings from the root
    Set ConfigRoot = EngineConfig.Root(True)
    
    With ConfigRoot
    
        ' Transfer all the configuration stuff to the structure
        
        ' Engine
        With .Item("engine")
            Engine.name = .Item("name")
            Engine.Version = .Item("version")
            Engine.ConfigFile = Mid$(filename, InStrRev(filename, "\") + 1)
            Engine.TableFile = .Item("table")
            Engine.AllowExeRead = .Item("allowexeread")
            Engine.ExeSize = .Item("exesize")
            Engine.Extended = .Item("extended")
        End With
        
        ' Item counts
        With .Item("itemcounts")
            ReDim Engine.Things(.Item("things"))
            ReDim Engine.ThingNames(.Item("things"))
            
            ReDim Engine.States(.Item("states"))
            ReDim Engine.CodePointers(.Item("states"))
            
            ReDim Engine.Sounds(.Item("sounds"))
            ReDim Engine.SoundNames(.Item("sounds"))
            
            ReDim Engine.SpriteNames(.Item("sprites"))
            ReDim Engine.SpritePointers(.Item("sprites"))
            
            ReDim Engine.Strings(.Item("strings"))
            
            ReDim Engine.Weapons(.Item("weapons"))
            ReDim Engine.WeaponNames(.Item("weapons"))
            
            ReDim Engine.Ammo(.Item("ammunition"))
            ReDim Engine.AmmoNames(.Item("ammunition"))
            
            ReDim Engine.Cheats(.Item("cheats"))
        End With
        
        ' Offsets
        If .Exists("offsets") Then
            With .Item("offsets")
                Engine.OffsetDataSegment = .Item("datasegment")
                Engine.OffsetThings = .Item("things")
                Engine.OffsetStates = .Item("states")
                Engine.OffsetSounds = .Item("sounds")
                Engine.OffsetSprites = .Item("sprites")
                Engine.OffsetStrings = .Item("strings")
                Engine.OffsetCodePointers = .Item("codepointers")
                Engine.OffsetWeapons = .Item("weapons")
                Engine.OffsetAmmo = .Item("ammunition")
                Engine.OffsetCheats = .Item("cheats")
            End With
        End If
        
        ' Flag names
        Set Engine.flags = .Item("flagnames")
        
        ' Flag tooltips
        Set Engine.FlagTooltips = .Item("flagtooltips")
        
        ' Thing names
        Set TempDic = .Item("thingnames")
        For A = 0 To TempDic.Count - 1
            Engine.ThingNames(A) = TempDic.Keys(A)
        Next A
        
        ' Weapon names
        Set TempDic = .Item("weaponnames")
        For A = 0 To TempDic.Count - 1
            Engine.WeaponNames(A) = TempDic.Items(A)
        Next A
        
        ' Ammo names
        Set TempDic = .Item("ammonames")
        For A = 0 To TempDic.Count - 1
            Engine.AmmoNames(A) = TempDic.Items(A)
        Next A
        
        ' Action descriptions
        Set Engine.ActionInfo = .Item("actiondescriptions")
        
        ' Codepointer mnemonics
        Set Engine.CodePointerMnemonics = .Item("codepointermnemonics")
        
        ' Cheat string offsets
        If .Exists("cheatoffsets") Then
            Set TempDic = .Item("cheatoffsets")
            For A = 0 To TempDic.Count - 1
                Engine.Cheats(A).offset = TempDic.Keys(A)
            Next A
        End If
        
        ' Cheat string lengths
        If .Exists("cheatlengths") Then
            Set TempDic = .Item("cheatlengths")
            For A = 0 To TempDic.Count - 1
                Engine.Cheats(A).length = TempDic.Items(A)
            Next A
        End If
        
        ' Codepointers to states
        If .Exists("codepointerstostates") Then
            Set TempDic = .Item("codepointerstostates")
            ReDim Engine.CodePointerToState(TempDic.Count - 1)
            For A = 0 To TempDic.Count - 1
                Engine.CodePointerToState(A) = TempDic.Keys(A)
            Next A
        End If
        
        ' Codepointer offsets
        If .Exists("codepointeroffsets") Then
            Set TempDic = .Item("codepointeroffsets")
            ReDim Engine.CodePointerOffsets(TempDic.Count - 1)
            For A = 0 To TempDic.Count - 1
                Engine.CodePointerOffsets(A) = TempDic.Keys(A)
            Next A
        End If
        
        ' Boom strings
        If .Exists("boomstrings") Then
            Set Engine.ExtendedStrings = .Item("boomstrings")
        Else
            Set Engine.ExtendedStrings = Nothing
        End If
        
        ' Misc. offsets (die, evil code!)
        If .Exists("miscellaneousoffsets") Then
            With .Item("miscellaneousoffsets")
                Engine.MiscOffsets.BFGUsage1 = .Item("bfgammousage1")
                Engine.MiscOffsets.BFGUsage2 = .Item("bfgammousage2")
                Engine.MiscOffsets.BFGUsage3 = .Item("bfgammousage3")
                Engine.MiscOffsets.BlueArmorClass = .Item("bluearmorclass")
                Engine.MiscOffsets.GodHealth1 = .Item("godmodehealth1")
                Engine.MiscOffsets.GodHealth2 = .Item("godmodehealth2")
                Engine.MiscOffsets.GreenArmorClass = .Item("greenarmorclass")
                Engine.MiscOffsets.IDFAArmor = .Item("idfaarmor")
                Engine.MiscOffsets.IDFAArmorClass = .Item("idfaarmorclass")
                Engine.MiscOffsets.IDKFAArmor = .Item("idkfaarmor")
                Engine.MiscOffsets.IDKFAArmorClass = .Item("idkfaarmorclass")
                Engine.MiscOffsets.InitialAmmo = .Item("initammo")
                Engine.MiscOffsets.InitialHealth = .Item("inithealth")
                Engine.MiscOffsets.MaxArmor1 = .Item("maxarmor1")
                Engine.MiscOffsets.MaxArmor2 = .Item("maxarmor2")
                Engine.MiscOffsets.MaxHealth1 = .Item("maxhealth1")
                Engine.MiscOffsets.MaxHealth2 = .Item("maxhealth2")
                Engine.MiscOffsets.MaxSoulHealth1 = .Item("maxsoulspherehealth1")
                Engine.MiscOffsets.MaxSoulHealth2 = .Item("maxsoulspherehealth2")
                Engine.MiscOffsets.MegaHealth = .Item("megaspherehealth")
                Engine.MiscOffsets.MonstersInfight = .Item("monstersinfight")
                Engine.MiscOffsets.SoulHealth = .Item("soulspherehealth")
            End With
        End If
        
    End With

End Function


' Read all engine configuration file from the \engines path
Public Sub GetEngineConfigs()
    
    Dim filename As String
    Dim ConfigClass As clsConfiguration
    Dim myFile As file
    
    
    ' Clear configuration info
    ConfigFiles.RemoveAll
    While ConfigFilesData.Count > 0
        ConfigFilesData.Remove 1
    Wend
    
    ' Find all engine configuration files
    For Each myFile In FS.GetFolder(APP_PATH & "config").Files
        
        filename = myFile.Path
        If LCase$(FS.GetExtensionName(filename)) = "cfg" Then
            
            ' Load a new engine configuration class
            Set ConfigClass = New clsConfiguration
            ConfigClass.LoadConfiguration filename
            
            ' Add the class if it is valid
            If ConfigClass.Root(False).Item("type") = CFGTYPE_ENGINE Then
                ConfigFiles.Add filename, ConfigClass.Root(False).Item("engine").Item("name")
                ConfigFilesData.Add ConfigClass.Root(False)("engine")("version")
            End If
            
            ' Destroy the class and get the next file
            Set ConfigClass = Nothing
            
        End If
        
    Next myFile

End Sub
