Attribute VB_Name = "modParse"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modParse.bas
'   Purpose: Small string-parsing routines.
'
Option Explicit
Option Compare Binary
Option Base 0


' Struct for Parse Data
Public Type CfgParseData
    Header As String
    Variable As String
    Value As String
    Kind As Byte
End Type

' Enum for config file parsing
Public Enum Parse
    psNone
    psHeader
    psVariable
End Enum


' Parse a line into separate arguments - console command style
' NOTE: Currently unused
Public Sub ParseLogLine(ByVal ParseString As String, ByRef Args() As String)

    Dim Pos As Long
    Dim Char As String
    Dim ArgUBound As Long
    Dim InString As Boolean


    ' Init the argument array
    ReDim Args(0)
    ArgUBound = 0

    ' Begin parsing
    For Pos = 1 To Len(ParseString)

        ' Get a character from the string
        Char = Mid(ParseString, Pos, 1)


        ' Check if we need to add an arg
        If Char = " " And InString = False Then

            ' Finish the current arg
            Args(ArgUBound) = Trim$(Args(ArgUBound))

            ' Redim the args array and andvance one arg
            ArgUBound = UBound(Args()) + 1
            ReDim Preserve Args(ArgUBound)

            ' Prevent the space from being added
            Char = vbNullString

        End If

        ' Check if we are starting a string
        If Char = Chr$(34) Then
            If InString = True Then InString = False Else InString = True
            Char = vbNullString
        End If


        ' Add the character
        Args(ArgUBound) = Args(ArgUBound) + Char

    Next Pos

End Sub


' Parse a line into separate arguments -- configuration file style
' NOTE: Currently unused
Public Sub ParseConfigLine(ByVal ParseString As String, ByRef VarData As CfgParseData)

    ' Trim the parse string and remove comments
    ParseString = Trim$(ParseString)
    If InStr(1, ParseString, "#") > 0 Then ParseString = Trim$(Left$(ParseString, InStr(1, ParseString, "#") - 1))

    ' Reset the struct
    VarData.Header = vbNullString
    VarData.Kind = psNone
    VarData.Value = vbNullString
    VarData.Variable = vbNullString

    ' Check if it is a header
    If Left$(ParseString, 1) = "[" And Right$(ParseString, 1) = "]" Then
        VarData.Kind = psHeader
        VarData.Header = Mid(ParseString, 2, Len(ParseString) - 2)
        Exit Sub
    End If

    ' Check if it is a variable
    If InStr(1, ParseString, "=") > 0 Then
        VarData.Kind = psVariable
        VarData.Variable = Trim$(Left$(ParseString, InStr(1, ParseString, "=") - 1))
        VarData.Value = Trim$(Right$(ParseString, Len(ParseString) - (InStr(1, ParseString, "=") + 1)))
        Exit Sub
    End If

End Sub


' Split a string into a Long type array (dynamically)
' NOTE: Currently unused
Public Sub SplitToLong(ByVal SplitString As String, ByVal Delimeter As String, ByRef ArrayData() As Long)

    Dim A As Long
    Dim Char As String
    Dim cIndex As Long
    Dim cValue As String


    ' Remove trailing delimeter
    SplitString = Trim$(SplitString)
    If Right$(SplitString, 1) = Delimeter Then SplitString = Trim$(Left$(SplitString, Len(SplitString) - 1))


    For A = 1 To Len(SplitString)

        ' Get current character
        Char = Mid$(SplitString, A, 1)

        ' See what we have
        If Char = Delimeter Then
            ' New indice
            ReDim Preserve ArrayData(cIndex)
            ArrayData(cIndex) = CLng(Trim$(cValue))

            ' Go to next indice
            cIndex = cIndex + 1
            cValue = vbNullString
        Else
            ' Add to current value
            cValue = cValue + Char
        End If

    Next A

    ' Add last indice too
    ReDim Preserve ArrayData(cIndex)
    ArrayData(cIndex) = CLng(Trim$(cValue))

End Sub


' Find simple cmdline commands
Public Function CommandLine(ByVal Search As String) As Boolean

    Search = LCase$(Search)
    
    If InStr(1, LCase$(Command), Search) > 0 Then CommandLine = True Else CommandLine = False

End Function


' Find used commandline parameters
Public Function CmdLine(ByVal Parameter As String) As String

    Dim A As Long
    Dim FindPos As Long
    
    
    FindPos = InStr(1, UCase$(Command), UCase$(Parameter))
    
    ' We found the parameter, see if there are any parameters to the parameter (...)
    If FindPos > 0 Then
        
        For A = FindPos + Len(Parameter) To Len(Command)
            
            ' We found a whole new parameter
            If Mid$(Command, A, 1) = "-" Then
                CmdLine = Trim$(LCase$(Mid$(Command, FindPos + Len(Parameter), A - FindPos + Len(Parameter))))
                Exit Function
            End If
            
        Next A
    
        CmdLine = Trim$(LCase$(Mid$(Command, FindPos + Len(Parameter))))
        Exit Function
        
    End If
    
    
    ' Return nothing
    CmdLine = vbNullString
    Exit Function
    
End Function
