Attribute VB_Name = "modPatchFiles"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modPatchFiles.bas
'   Purpose: Saving and loading of dehacked\bex files.
'
Option Explicit
Option Compare Binary
Option Base 0


' Type of detected line
Private Enum LineTypeEnum
    lnNone = 0
    lnAssignment
    lnHeader
End Enum

' Type of current deh block
Private Enum StructTypeEnum
    strThing = 0
    strState
    strSound
    strCodePointer
    strWeapon
    strAmmo
    strCheat
    strMisc
    strString
    strPars
    strExtCodePointer
    strExtStrings
End Enum


' The string buffer
Private StringBuffer As String


' Save a patch file
Public Sub SavePatch(ByVal filename As String)

    Dim A As Long
    Dim B As Long
    Dim C As Long
    Dim openFile As Long
    Dim EntryName As String
    Dim Wrote As Boolean


    ' Make a backup if the file exists first
    If Config.Root(False)("createBackup") = 1 Then
        If FS.FileExists(filename) Then FileCopy filename, filename & ".bak"
    End If


    openFile = FreeFile
    Open filename For Output As #openFile

        ' Write the header
        Print #openFile, "Patch File for DeHackEd v3.0"
        Print #openFile, "# Created with " & VersionString
        Print #openFile, ""

        ' Write the version info
        Print #openFile, "# General version information"
        Print #openFile, "Doom version = " & Patch.EngineVersion
        Print #openFile, "Patch format = " & Patch.Version
        Print #openFile, ""
        
        ' Write WhackEd3 specific info
        Print #openFile, "# Info for WhackEd3"
        Print #openFile, "Engine config = " & Engine.ConfigFile
        If Not (dataWad Is Nothing) Then Print #openFile, "Data WAD = " & dataWad.WadFileName
        Print #openFile, "IWAD = " & mainWad.WadFileName
        Print #openFile, ""
        Print #openFile, ""


        ' Start writing thing table
        Wrote = False
        For A = 0 To UBound(Patch.Things())
            StringBuffer = vbNullString

            ' See if the name changed too
            If Patch.ThingNames(A) <> Engine.ThingNames(A) Then
                EntryName = Patch.ThingNames(A)
            Else
                EntryName = Engine.ThingNames(A)
            End If

            With Patch.Things(A)
                If .id <> Engine.Things(A).id Then AddBuffer "ID # = " & .id
                If .SpawnState <> Engine.Things(A).SpawnState Then AddBuffer "Initial frame = " & .SpawnState
                If .HitPoints <> Engine.Things(A).HitPoints Then AddBuffer "Hit points = " & .HitPoints
                If .MoveState <> Engine.Things(A).MoveState Then AddBuffer "First moving frame = " & .MoveState
                If .SeePlayerSnd <> Engine.Things(A).SeePlayerSnd Then AddBuffer "Alert sound = " & .SeePlayerSnd
                If .ReactionTime <> Engine.Things(A).ReactionTime Then AddBuffer "Reaction time = " & .ReactionTime
                If .AttackSnd <> Engine.Things(A).AttackSnd Then AddBuffer "Attack sound = " & .AttackSnd
                If .PainState <> Engine.Things(A).PainState Then AddBuffer "Injury frame = " & .PainState
                If .PainChance <> Engine.Things(A).PainChance Then AddBuffer "Pain chance = " & .PainChance
                If .PainSnd <> Engine.Things(A).PainSnd Then AddBuffer "Pain sound = " & .PainSnd
                If .MeleeState <> Engine.Things(A).MeleeState Then AddBuffer "Close attack frame = " & .MeleeState
                If .AttackState <> Engine.Things(A).AttackState Then AddBuffer "Far attack frame = " & .AttackState
                If .DeathState <> Engine.Things(A).DeathState Then AddBuffer "Death frame = " & .DeathState
                If .ExplodingState <> Engine.Things(A).ExplodingState Then AddBuffer "Exploding frame = " & .ExplodingState
                If .DeathSnd <> Engine.Things(A).DeathSnd Then AddBuffer "Death sound = " & .DeathSnd
                If .Speed <> Engine.Things(A).Speed Then AddBuffer "Speed = " & .Speed
                If .Radius <> Engine.Things(A).Radius Then AddBuffer "Width = " & .Radius
                If .Height <> Engine.Things(A).Height Then AddBuffer "Height = " & .Height
                If .Mass <> Engine.Things(A).Mass Then AddBuffer "Mass = " & .Mass
                If .MissileDamage <> Engine.Things(A).MissileDamage Then AddBuffer "Missile damage = " & .MissileDamage
                If .ActSnd <> Engine.Things(A).ActSnd Then AddBuffer "Action sound = " & .ActSnd
                
                ' Flags, handled for extended engine and signs in another function
                If .flags <> Engine.Things(A).flags Then AddBuffer "Bits = " & ExtendedThingFlags(.flags)
                
                If .RespawnState <> Engine.Things(A).RespawnState Then AddBuffer "Respawn frame = " & .RespawnState
            End With

            If StringBuffer <> vbNullString Or EntryName <> Engine.ThingNames(A) Then
                Print #openFile, "Thing " & A + 1 & " (" & EntryName & ")"
                Print #openFile, StringBuffer
                Wrote = True
            End If
        Next A
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If


        ' Start writing state table
        Wrote = False
        For A = 0 To UBound(Patch.States())
            StringBuffer = vbNullString
            
            With Patch.States(A)
                If .Sprite <> Engine.States(A).Sprite Then AddBuffer "Sprite number = " & .Sprite
                If .Frame <> Engine.States(A).Frame Then AddBuffer "Sprite subnumber = " & .Frame
                If .Duration <> Engine.States(A).Duration Then AddBuffer "Duration = " & .Duration
                If .NextState <> Engine.States(A).NextState Then AddBuffer "Next frame = " & .NextState
                If .Parameter1 <> Engine.States(A).Parameter1 Then AddBuffer "Unknown 1 = " & .Parameter1
                If .Parameter2 <> Engine.States(A).Parameter2 Then AddBuffer "Unknown 2 = " & .Parameter2
            End With

            If StringBuffer <> vbNullString Then
                Print #openFile, "Frame " & A
                Print #openFile, StringBuffer
                Wrote = True
            End If
        Next A
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If


        ' Start writing sound table
        Wrote = False
        For A = 0 To UBound(Patch.Sounds())
            StringBuffer = vbNullString

            With Patch.Sounds(A)
                If .StringPointer <> Engine.Sounds(A).StringPointer Then AddBuffer "Offset = " & .StringPointer
                If .Singularity <> Engine.Sounds(A).Singularity Then AddBuffer "Zero/One = " & .Singularity
                If .priority <> Engine.Sounds(A).priority Then AddBuffer "Value = " & .priority
                If .Reference <> Engine.Sounds(A).Reference Then AddBuffer "Zero 1 = " & .Reference
                If .Pitch <> Engine.Sounds(A).Pitch Then AddBuffer "Neg. One 1 = " & .Pitch
                If .volume <> Engine.Sounds(A).volume Then AddBuffer "Neg. One 2 = " & .volume
                If .data <> Engine.Sounds(A).data Then AddBuffer "Zero 2 = " & .data
                If .Usefulness <> Engine.Sounds(A).Usefulness Then AddBuffer "Zero 3 = " & .Usefulness
                If .LumpNum <> Engine.Sounds(A).LumpNum Then AddBuffer "Zero 4 = " & .LumpNum
            End With

            If StringBuffer <> vbNullString Then
                Print #openFile, "Sound " & A + 1
                Print #openFile, StringBuffer
                Wrote = True
            End If
        Next A
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If


        ' Start writing the codepointers
        If Engine.Extended = False Then

            Wrote = False
            For A = 0 To UBound(Patch.CodePointers())
    
                If Patch.CodePointers(A) <> Engine.CodePointers(A) Then
    
                    ' Find a matching state
                    For B = 0 To UBound(Patch.States())
                        If Patch.States(B).CodePointer = Patch.CodePointers(A) Then Exit For
                    Next B
    
                    ' Find a codepointer matching the frame
                    For C = 0 To UBound(Engine.CodePointerToState())
                        If Engine.CodePointerToState(C) = A Then Exit For
                    Next C
    
                    ' Write the codepointer
                    If B <> UBound(Patch.States()) Then
                        Print #openFile, "Pointer " & C & " (Frame " & A & ")"
                        Print #openFile, "Codep Frame = " & B
                        Print #openFile, ""
                        Wrote = True
                    End If
    
                End If
    
            Next A
        
        ' Write extended codepointers
        Else
        
            StringBuffer = vbNullString
            Wrote = False
            For A = 0 To UBound(Engine.CodePointers())
                
                If Patch.CodePointers(A) <> Engine.CodePointers(A) Then
                    AddBuffer "Frame " & Trim$(Str$(A)) & " = " & Engine.CodePointerMnemonics.Item(Trim$(Str$(Patch.CodePointers(A))))
                End If
            
            Next A
            
            ' Write string buffer
            If StringBuffer <> vbNullString Then
                Print #openFile, "[CODEPTR]"
                Print #openFile, StringBuffer
                Wrote = True
            End If
            
        End If
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If
        
        
        ' Start writing weapons
        Wrote = False
        For A = 0 To UBound(Patch.WeaponNames())
            StringBuffer = vbNullString

            ' See if the name changed too
            If Patch.WeaponNames(A) <> Engine.WeaponNames(A) Then
                EntryName = Patch.WeaponNames(A)
            Else
                EntryName = Engine.WeaponNames(A)
            End If

            With Patch.Weapons(A)
                If .AmmoType <> Engine.Weapons(A).AmmoType Then AddBuffer "Ammo type = " & .AmmoType
                If .DeselectState <> Engine.Weapons(A).DeselectState Then AddBuffer "Deselect frame = " & .DeselectState
                If .SelectState <> Engine.Weapons(A).SelectState Then AddBuffer "Select frame = " & .SelectState
                If .BobbingState <> Engine.Weapons(A).BobbingState Then AddBuffer "Bobbing frame = " & .BobbingState
                If .ShootingState <> Engine.Weapons(A).ShootingState Then AddBuffer "Shooting frame = " & .ShootingState
                If .MuzzleState <> Engine.Weapons(A).MuzzleState Then AddBuffer "Firing frame = " & .MuzzleState
            End With

            If StringBuffer <> vbNullString Or EntryName <> Engine.WeaponNames(A) Then
                Print #openFile, "Weapon " & A & " (" & EntryName & ")"
                Print #openFile, StringBuffer
                Wrote = True
            End If
        
        Next A
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If
        
        
        ' Start writing ammo
        Wrote = False
        For A = 0 To UBound(Patch.AmmoNames())
            StringBuffer = vbNullString

            ' See if the name changed too
            If Patch.AmmoNames(A) <> Engine.AmmoNames(A) Then
                EntryName = Patch.AmmoNames(A)
            Else
                EntryName = Engine.AmmoNames(A)
            End If

            With Patch.Ammo(A)
                If .Maximum <> Engine.Ammo(A).Maximum Then AddBuffer "Max ammo = " & .Maximum
                If .Pickup <> Engine.Ammo(A).Pickup Then AddBuffer "Per ammo = " & .Pickup
            End With

            If StringBuffer <> vbNullString Or EntryName <> Engine.AmmoNames(A) Then
                Print #openFile, "Ammo " & A & " (" & EntryName & ")"
                Print #openFile, StringBuffer
                Wrote = True
            End If
        
        Next A
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If
        
        
        ' Write the cheat strings
        StringBuffer = vbNullString
        With Patch
            If .Cheats(0).Text <> Engine.Cheats(0).Text Then AddBuffer "Change music = " & Pad255(.Cheats(0).Text, .Cheats(0).length)
            If .Cheats(1).Text <> Engine.Cheats(1).Text Then AddBuffer "Chainsaw = " & Pad255(.Cheats(1).Text, .Cheats(1).length)
            If .Cheats(2).Text <> Engine.Cheats(2).Text Then AddBuffer "God mode = " & Pad255(.Cheats(2).Text, .Cheats(2).length)
            If .Cheats(3).Text <> Engine.Cheats(3).Text Then AddBuffer "Ammo & Keys = " & Pad255(.Cheats(3).Text, .Cheats(3).length)
            If .Cheats(4).Text <> Engine.Cheats(4).Text Then AddBuffer "Ammo = " & Pad255(.Cheats(4).Text, .Cheats(4).length)
            If .Cheats(5).Text <> Engine.Cheats(5).Text Then AddBuffer "No Clipping 1 = " & Pad255(.Cheats(5).Text, .Cheats(5).length)
            If .Cheats(6).Text <> Engine.Cheats(6).Text Then AddBuffer "No Clipping 2 = " & Pad255(.Cheats(6).Text, .Cheats(6).length)
            If .Cheats(7).Text <> Engine.Cheats(7).Text Then AddBuffer "Invincibility = " & Pad255(.Cheats(7).Text, .Cheats(7).length)
            If .Cheats(8).Text <> Engine.Cheats(8).Text Then AddBuffer "Berserk = " & Pad255(.Cheats(8).Text, .Cheats(8).length)
            If .Cheats(9).Text <> Engine.Cheats(9).Text Then AddBuffer "Invisibility = " & Pad255(.Cheats(9).Text, .Cheats(9).length)
            If .Cheats(10).Text <> Engine.Cheats(10).Text Then AddBuffer "Radiation Suit = " & Pad255(.Cheats(10).Text, .Cheats(10).length)
            If .Cheats(11).Text <> Engine.Cheats(11).Text Then AddBuffer "Auto-map = " & Pad255(.Cheats(11).Text, .Cheats(11).length)
            If .Cheats(12).Text <> Engine.Cheats(12).Text Then AddBuffer "Lite-Amp Goggles = " & Pad255(.Cheats(12).Text, .Cheats(12).length)
            If .Cheats(13).Text <> Engine.Cheats(13).Text Then AddBuffer "BEHOLD menu = " & Pad255(.Cheats(13).Text, .Cheats(13).length)
            If .Cheats(14).Text <> Engine.Cheats(14).Text Then AddBuffer "Level Warp = " & Pad255(.Cheats(14).Text, .Cheats(14).length)
            If .Cheats(15).Text <> Engine.Cheats(15).Text Then AddBuffer "Player Position = " & Pad255(.Cheats(15).Text, .Cheats(15).length)
            If .Cheats(16).Text <> Engine.Cheats(16).Text Then AddBuffer "Map cheat = " & Pad255(.Cheats(16).Text, .Cheats(16).length)
        End With
        
        If StringBuffer <> vbNullString Then
            Print #openFile, ""
            Print #openFile, ""
            Print #openFile, "Cheat 0"
            Print #openFile, StringBuffer
        End If
        
        
        ' Start writing misc data
        StringBuffer = vbNullString
        
        With Patch.Misc
            If .InitialHealth <> Engine.Misc.InitialHealth Then AddBuffer "Initial Health = " & .InitialHealth
            If .InitialAmmo <> Engine.Misc.InitialAmmo Then AddBuffer "Initial Bullets = " & .InitialAmmo
            If .MaxHealth1 <> Engine.Misc.MaxHealth1 Then AddBuffer "Max Health = " & .MaxHealth1
            If .MaxArmor1 <> Engine.Misc.MaxArmor1 Then AddBuffer "Max Armor = " & .MaxArmor1
            If .GreenArmorClass <> Engine.Misc.GreenArmorClass Then AddBuffer "Green Armor Class = " & .GreenArmorClass
            If .BlueArmorClass <> Engine.Misc.BlueArmorClass Then AddBuffer "Blue Armor Class = " & .BlueArmorClass
            If .MaxSoulHealth1 <> Engine.Misc.MaxSoulHealth1 Then AddBuffer "Max Soulsphere = " & .MaxSoulHealth1
            If .SoulHealth <> Engine.Misc.SoulHealth Then AddBuffer "Soulsphere Health = " & .SoulHealth
            If .MegaHealth <> Engine.Misc.MegaHealth Then AddBuffer "Megasphere Health = " & .MegaHealth
            If .GodHealth1 <> Engine.Misc.GodHealth1 Then AddBuffer "God Mode Health = " & .GodHealth1
            If .IDFAArmor <> Engine.Misc.IDFAArmor Then AddBuffer "IDFA Armor = " & .IDFAArmor
            If .IDFAArmorClass <> Engine.Misc.IDFAArmorClass Then AddBuffer "IDFA Armor Class = " & .IDFAArmorClass
            If .IDKFAArmor <> Engine.Misc.IDKFAArmor Then AddBuffer "IDKFA Armor = " & .IDKFAArmor
            If .IDKFAArmorClass <> Engine.Misc.IDKFAArmorClass Then AddBuffer "IDKFA Armor Class = " & .IDKFAArmorClass
            If .BFGUsage1 <> Engine.Misc.BFGUsage1 Then AddBuffer "BFG Cells/Shot = " & .BFGUsage1
            If .MonstersInfight <> Engine.Misc.MonstersInfight Then AddBuffer "Monsters Infight = " & .MonstersInfight
        End With
        
        If StringBuffer <> vbNullString Then
            Print #openFile, ""
            Print #openFile, ""
            Print #openFile, "Misc 0"
            Print #openFile, StringBuffer
        End If
        
        
        
        ' Write par. times
        StringBuffer = vbNullString
        
        For A = 0 To Patch.nPars - 1
            ' MAPxy
            If Patch.Pars(A).Episode = 0 Then
                AddBuffer "par " & Trim$(Str$(Patch.Pars(A).Map)) & " " & Trim$(Str$(Patch.Pars(A).time))
            
            ' ExMy
            Else
                AddBuffer "par " & Trim$(Str$(Patch.Pars(A).Episode)) & " " & Trim$(Str$(Patch.Pars(A).Map)) & " " & Trim$(Str$(Patch.Pars(A).time))
                
            End If
        Next A
        
        If StringBuffer <> vbNullString Then
            Print #openFile, "[PARS]"
            Print #openFile, StringBuffer
            Wrote = True
        End If
        
        If Wrote = True Then
            Print #openFile, ""
            Print #openFile, ""
        End If
        
        
        ' Start writing strings (note that these should ALWAYS go last)
        If Engine.Extended = False Then

            For A = 0 To UBound(Patch.Strings())
            
                ' Write string if changed
                If Patch.Strings(A).Text <> Engine.Strings(A).Text Then
                    Print #openFile, "Text " & Engine.Strings(A).length & " " & Patch.Strings(A).length
                    Print #openFile, Engine.Strings(A).Text & Patch.Strings(A).Text
                    Print #openFile, ""
                End If
            
            Next A
        Else
        
            ' Cycle through each string
            StringBuffer = vbNullString
            For A = 0 To Engine.ExtendedStrings.Count - 1
                
                ' Add to string buffer if changed
                If Patch.Strings(A).Text <> Engine.ExtendedStrings.Items(A) Then
                    AddBuffer Engine.ExtendedStrings.Keys(A) & " = " & Escaped(Patch.Strings(A).Text)
                End If
                
            Next A
            
            ' Write the string buffer if anything is in it
            If StringBuffer <> vbNullString Then
                Print #openFile, ""
                Print #openFile, ""
                Print #openFile, "[STRINGS]"
                Print #openFile, StringBuffer
            End If
        
        End If
        
        
    Close #openFile

    appLog.Add "Saved " & filename

End Sub


' Add a string to the string buffer, a small but useful wrapper
Private Sub AddBuffer(ByVal AddString As String)

    StringBuffer = StringBuffer & AddString & vbCrLf

End Sub


' Convert to extended thing flags
Private Function ExtendedThingFlags(ByVal flags As Long) As String

    Dim A As Long
    Dim Result As String
    
    
    ' Return default if 0
    If flags = 0 Then
        ExtendedThingFlags = "0"
        Exit Function
    End If
    
    ' Do nothing if engine is not extended
    If Engine.Extended = False Then
        If flags < 0 Then
            ExtendedThingFlags = Trim$(Str$(CVar(flags + (2 ^ 32))))
        Else
            ExtendedThingFlags = Trim$(Str$(flags))
        End If
        
        Exit Function
    End If
    
    
    ' Check for every flag
    Result = vbNullString
    For A = 0 To Engine.flags.Count - 1
        If GetFlag(flags, A) = 1 Then Result = Result & Engine.flags.Keys(A) & "+"
    Next A
    
    ' Remove eventual trailing +
    If Right$(Result, 1) = "+" Then Result = Mid$(Result, 1, Len(Result) - 1)
    
    ' Return the result
    ExtendedThingFlags = Result

End Function


' Load a patch file
Public Function LoadPatch(ByRef filename As String) As Boolean

    Dim A As Long
    Dim Unsupported As Boolean

    Dim openFile As Long

    Dim ReadLine As String
    Dim ValidFile As Boolean
    Dim ValidEngine As Boolean
    Dim ValidFormat As Boolean
    Dim inString As Boolean

    Dim Args() As String
    Dim cWord As String
    Dim Char As String

    Dim LineType As Byte
    Dim Entry As Long
    Dim StructType As StructTypeEnum
    
    Dim CPFrameNum As Long
    
    ' Text string parsing
    Dim OldString As String
    Dim NewString As String
    Dim OldLen As Long
    Dim NewLen As Long
    Dim Space1 As Long
    Dim Space2 As Long
    Dim Replacement As Boolean


    ' Reset patch data
    modTables.ResetPatch

    ' Default values
    ValidFile = False
    ValidFormat = False
    ValidEngine = False

    ' Start loading the file
    openFile = FreeFile
    Open filename For Input As #openFile

        Do

            ' Read a single line
            Line Input #openFile, ReadLine
            ReadLine = Trim$(ReadLine)


            ' Quick checks for right file type
            If ReadLine = "Patch File for DeHackEd v3.0" Then
                ValidFile = True
                ReadLine = vbNullString
            End If

            ' See if the file is valid at all
            If ValidFile = False Then
                MsgBox "Invalid patch file.", vbExclamation Or vbOKOnly
                LoadPatch = False
                Exit Function
            End If


            ' Parse the line into useable arguments
            If InStr(1, ReadLine, "=") Then

                ' We have a variable\value combination
                ReDim Args(1)
                Args(0) = Trim$(Left$(ReadLine, InStr(1, ReadLine, "=") - 1))
                Args(1) = Trim$(Mid$(ReadLine, InStr(1, ReadLine, "=") + 1))
                LineType = lnAssignment

            Else

                ' Just parse the line into args separated by spaces or special chars
                ReDim Args(0)
                cWord = vbNullString
                For A = 1 To Len(ReadLine)

                    Char = Mid$(ReadLine, A, 1)

                    ' A space separator
                    If Char = Chr$(32) And inString = False Then
                        Args(UBound(Args())) = Trim$(cWord)
                        ReDim Preserve Args(UBound(Args()) + 1)
                        Char = vbNullString
                        cWord = vbNullString
                    End If

                    ' Start of a string
                    If Char = "(" Then
                        inString = True
                        Char = vbNullString
                    End If

                    ' End of a string
                    If Char = ")" Then
                        inString = False
                        Char = vbNullString
                    End If

                    ' Comments - ignore the rest of the line
                    If Char = "#" Then
                        cWord = vbNullString
                        Exit For
                    End If

                    cWord = cWord & Char

                Next A

                ' Add the last word if any
                If Trim$(cWord) <> vbNullString Then
                    Args(UBound(Args())) = Trim$(cWord)
                    ReDim Preserve Args(UBound(Args()) + 1)
                End If


                ' Determine the structure's type that we are inside of
                Select Case UCase$(Args(0))

                    Case "THING"
                        StructType = strThing
                        Entry = Val(Args(1)) - 1
                        Patch.ThingNames(Entry) = Args(2)
                        LineType = lnHeader

                    Case "FRAME"
                        StructType = strState
                        Entry = Val(Args(1))
                        LineType = lnHeader

                    Case "SOUND"
                        StructType = strSound
                        Entry = Val(Args(1))
                        LineType = lnHeader

                    Case "POINTER"
                        StructType = strCodePointer
                        Entry = Val(Args(1))
                        LineType = lnHeader
                        
                    Case "WEAPON"
                        StructType = strWeapon
                        Entry = Val(Args(1))
                        Patch.WeaponNames(Entry) = Args(2)
                        LineType = lnHeader
                        
                    Case "AMMO"
                        StructType = strAmmo
                        Entry = Val(Args(1))
                        Patch.AmmoNames(Entry) = Args(2)
                        LineType = lnHeader
                    
                    Case "TEXT"
                        StructType = strString
                        Entry = Val(Args(1))
                        LineType = lnAssignment
                        
                    Case "CHEAT"
                        StructType = strCheat
                        Entry = 0
                        LineType = lnHeader
                    
                    Case "MISC"
                        StructType = strMisc
                        Entry = 0
                        LineType = lnHeader
                    
                    Case "[CODEPTR]"
                        StructType = strExtCodePointer
                        Entry = 0
                        LineType = lnHeader
                    
                    Case "[STRINGS]"
                        StructType = strExtStrings
                        Entry = 0
                        LineType = lnHeader
                        
                    Case "[PARS]"
                        StructType = strPars
                        Entry = 0
                        LineType = lnHeader

                    
                    ' Unsupported stuff (these will *never* be supported, too messy)
                    Case "SPRITE", "INCLUDE"
                        Unsupported = True
                    
                End Select

            End If


            ' Par times don't use assignments, so they're placed here.
            If StructType = strPars Then
                If UCase$(Args(0)) = "PAR" Then
                    ExpandParList
                    
                    ' MAPxy
                    If UBound(Args()) = 3 Then
                        Patch.Pars(Patch.nPars - 1).Episode = 0
                        Patch.Pars(Patch.nPars - 1).Map = Val(Args(1))
                        Patch.Pars(Patch.nPars - 1).time = Val(Args(2))
                    
                    ' ExMy
                    ElseIf UBound(Args()) = 4 Then
                        Patch.Pars(Patch.nPars - 1).Episode = Val(Args(1))
                        Patch.Pars(Patch.nPars - 1).Map = Val(Args(2))
                        Patch.Pars(Patch.nPars - 1).time = Val(Args(3))
                    
                    End If
                    
                End If
            End If


            ' See what we have to change inside the table
            If LineType = lnAssignment Then

                Args(0) = UCase$(Args(0))


                ' Get version info
                If UBound(Args()) = 1 Then

                    ' Engine version info
                    If Args(0) = "DOOM VERSION" And Val(Args(1)) <> Engine.Version Then
                        MsgBox "Patch file version does not match engine version.", vbExclamation Or vbOKOnly
                        LoadPatch = False
                        Exit Function
                    ElseIf Args(0) = "DOOM VERSION" Then
                        ValidEngine = True
                    End If

                    ' Patch version info
                    If Args(0) = "PATCH FORMAT" And Val(Args(1)) <> 6 Then
                        MsgBox "Unsupported patch format.", vbExclamation Or vbOKOnly
                        LoadPatch = False
                        Exit Function
                    ElseIf Args(0) = "PATCH FORMAT" Then
                        ValidFormat = True
                    End If

                End If


                ' Don't read anything until we have confiarmation of a proper patch file
                If ValidEngine = True And ValidFormat = True Then

                    ' See what kind of table we are going to change
                    Select Case StructType

                        ' A thing
                        Case strThing
                            If Args(0) = "ID #" Then Patch.Things(Entry).id = Val(Args(1))
                            If Args(0) = "INITIAL FRAME" Then Patch.Things(Entry).SpawnState = Val(Args(1))
                            If Args(0) = "HIT POINTS" Then Patch.Things(Entry).HitPoints = Val(Args(1))
                            If Args(0) = "FIRST MOVING FRAME" Then Patch.Things(Entry).MoveState = Val(Args(1))
                            If Args(0) = "ALERT SOUND" Then Patch.Things(Entry).SeePlayerSnd = Val(Args(1))
                            If Args(0) = "REACTION TIME" Then Patch.Things(Entry).ReactionTime = Val(Args(1))
                            If Args(0) = "ATTACK SOUND" Then Patch.Things(Entry).AttackSnd = Val(Args(1))
                            If Args(0) = "INJURY FRAME" Then Patch.Things(Entry).PainState = Val(Args(1))
                            If Args(0) = "PAIN CHANCE" Then Patch.Things(Entry).PainChance = Val(Args(1))
                            If Args(0) = "PAIN SOUND" Then Patch.Things(Entry).PainSnd = Val(Args(1))
                            If Args(0) = "CLOSE ATTACK FRAME" Then Patch.Things(Entry).MeleeState = Val(Args(1))
                            If Args(0) = "FAR ATTACK FRAME" Then Patch.Things(Entry).AttackState = Val(Args(1))
                            If Args(0) = "DEATH FRAME" Then Patch.Things(Entry).DeathState = Val(Args(1))
                            If Args(0) = "EXPLODING FRAME" Then Patch.Things(Entry).ExplodingState = Val(Args(1))
                            If Args(0) = "DEATH SOUND" Then Patch.Things(Entry).DeathSnd = Val(Args(1))
                            If Args(0) = "SPEED" Then Patch.Things(Entry).Speed = Val(Args(1))
                            If Args(0) = "WIDTH" Then Patch.Things(Entry).Radius = Val(Args(1))
                            If Args(0) = "HEIGHT" Then Patch.Things(Entry).Height = Val(Args(1))
                            If Args(0) = "MASS" Then Patch.Things(Entry).Mass = Val(Args(1))
                            If Args(0) = "MISSILE DAMAGE" Then Patch.Things(Entry).MissileDamage = Val(Args(1))
                            If Args(0) = "ACTION SOUND" Then Patch.Things(Entry).ActSnd = Val(Args(1))
                            If Args(0) = "BITS" Then Patch.Things(Entry).flags = GetThingFlags(Args(1))
                            If Args(0) = "RESPAWN FRAME" Then Patch.Things(Entry).RespawnState = Val(Args(1))

                        ' A state
                        Case strState
                            If Args(0) = "SPRITE NUMBER" Then Patch.States(Entry).Sprite = Val(Args(1))
                            If Args(0) = "SPRITE SUBNUMBER" Then Patch.States(Entry).Frame = Val(Args(1))
                            If Args(0) = "DURATION" Then Patch.States(Entry).Duration = Val(Args(1))
                            If Args(0) = "NEXT FRAME" Then Patch.States(Entry).NextState = Val(Args(1))
                            If Args(0) = "UNKNOWN 1" Then Patch.States(Entry).Parameter1 = Val(Args(1))
                            If Args(0) = "UNKNOWN 2" Then Patch.States(Entry).Parameter2 = Val(Args(1))

                        ' A sound
                        Case strSound
                            If Args(0) = "OFFSET" Then Patch.Sounds(Entry).StringPointer = Val(Args(1))
                            If Args(0) = "ZERO/ONE" Then Patch.Sounds(Entry).Singularity = Val(Args(1))
                            If Args(0) = "VALUE" Then Patch.Sounds(Entry).priority = Val(Args(1))
                            If Args(0) = "ZERO 1" Then Patch.Sounds(Entry).Reference = Val(Args(1))
                            If Args(0) = "ZERO 2" Then Patch.Sounds(Entry).data = Val(Args(1))
                            If Args(0) = "ZERO 3" Then Patch.Sounds(Entry).Usefulness = Val(Args(1))
                            If Args(0) = "ZERO 4" Then Patch.Sounds(Entry).LumpNum = Val(Args(1))
                            If Args(0) = "NEG. ONE 1" Then Patch.Sounds(Entry).Pitch = Val(Args(1))
                            If Args(0) = "NEG. ONE 2" Then Patch.Sounds(Entry).volume = Val(Args(1))

                        ' A code pointer
                        Case strCodePointer
                            If Args(0) = "CODEP FRAME" Then Patch.CodePointers(Engine.CodePointerToState(Entry)) = Patch.States(Val(Args(1))).CodePointer
                            
                        ' Extended codepointer
                        Case strExtCodePointer
                            If Left$(Args(0), 6) = "FRAME " Then
                                CPFrameNum = Val(Trim$(Mid$(Args(0), 7)))
                                For A = 0 To Engine.CodePointerMnemonics.Count - 1
                                    If Engine.CodePointerMnemonics.Items(A) = Args(1) Then
                                        Patch.CodePointers(CPFrameNum) = Engine.CodePointerMnemonics.Keys(A)
                                        Exit For
                                    End If
                                Next A
                            End If
                            
                        ' A weapon
                        Case strWeapon
                            If Args(0) = "AMMO TYPE" Then Patch.Weapons(Entry).AmmoType = Val(Args(1))
                            If Args(0) = "DESELECT FRAME" Then Patch.Weapons(Entry).DeselectState = Val(Args(1))
                            If Args(0) = "SELECT FRAME" Then Patch.Weapons(Entry).SelectState = Val(Args(1))
                            If Args(0) = "BOBBING FRAME" Then Patch.Weapons(Entry).BobbingState = Val(Args(1))
                            If Args(0) = "SHOOTING FRAME" Then Patch.Weapons(Entry).ShootingState = Val(Args(1))
                            If Args(0) = "FIRING FRAME" Then Patch.Weapons(Entry).MuzzleState = Val(Args(1))
                        
                        ' An ammo entry
                        Case strAmmo
                            If Args(0) = "MAX AMMO" Then Patch.Ammo(Entry).Maximum = Val(Args(1))
                            If Args(0) = "PER AMMO" Then Patch.Ammo(Entry).Pickup = Val(Args(1))
                            
                        ' A cheat entry
                        Case strCheat
                            If Args(0) = "CHANGE MUSIC" Then Patch.Cheats(0).Text = Strip255(Args(1))
                            If Args(0) = "CHAINSAW" Then Patch.Cheats(1).Text = Strip255(Args(1))
                            If Args(0) = "GOD MODE" Then Patch.Cheats(2).Text = Strip255(Args(1))
                            If Args(0) = "AMMO & KEYS" Then Patch.Cheats(3).Text = Strip255(Args(1))
                            If Args(0) = "AMMO" Then Patch.Cheats(4).Text = Strip255(Args(1))
                            If Args(0) = "NO CLIPPING 1" Then Patch.Cheats(5).Text = Strip255(Args(1))
                            If Args(0) = "NO CLIPPING 2" Then Patch.Cheats(6).Text = Strip255(Args(1))
                            If Args(0) = "INVINCIBILITY" Then Patch.Cheats(7).Text = Strip255(Args(1))
                            If Args(0) = "BERSERK" Then Patch.Cheats(8).Text = Strip255(Args(1))
                            If Args(0) = "INVISIBILITY" Then Patch.Cheats(9).Text = Strip255(Args(1))
                            If Args(0) = "RADIATION SUIT" Then Patch.Cheats(10).Text = Strip255(Args(1))
                            If Args(0) = "AUTO-MAP" Then Patch.Cheats(11).Text = Strip255(Args(1))
                            If Args(0) = "LITE-AMP GOGGLES" Then Patch.Cheats(12).Text = Strip255(Args(1))
                            If Args(0) = "BEHOLD MENU" Then Patch.Cheats(13).Text = Strip255(Args(1))
                            If Args(0) = "LEVEL WARP" Then Patch.Cheats(14).Text = Strip255(Args(1))
                            If Args(0) = "PLAYER POSITION" Then Patch.Cheats(15).Text = Strip255(Args(1))
                            If Args(0) = "MAP CHEAT" Then Patch.Cheats(16).Text = Strip255(Args(1))
                            
                        ' A misc. entry
                        Case strMisc
                            If Args(0) = "INITIAL HEALTH" Then Patch.Misc.InitialHealth = Val(Args(1))
                            If Args(0) = "INITIAL BULLETS" Then Patch.Misc.InitialAmmo = Val(Args(1))
                            If Args(0) = "MAX HEALTH" Then Patch.Misc.MaxHealth1 = Val(Args(1))
                            If Args(0) = "MAX ARMOR" Then Patch.Misc.MaxArmor1 = Val(Args(1))
                            If Args(0) = "GREEN ARMOR CLASS" Then Patch.Misc.GreenArmorClass = Val(Args(1))
                            If Args(0) = "BLUE ARMOR CLASS" Then Patch.Misc.BlueArmorClass = Val(Args(1))
                            If Args(0) = "MAX SOULSPHERE" Then Patch.Misc.MaxSoulHealth1 = Val(Args(1))
                            If Args(0) = "SOULSPHERE HEALTH" Then Patch.Misc.SoulHealth = Val(Args(1))
                            If Args(0) = "MEGASPHERE HEALTH" Then Patch.Misc.MegaHealth = Val(Args(1))
                            If Args(0) = "GOD MODE HEALTH" Then Patch.Misc.GodHealth1 = Val(Args(1))
                            If Args(0) = "IDFA ARMOR" Then Patch.Misc.IDFAArmor = Val(Args(1))
                            If Args(0) = "IDFA ARMOR CLASS" Then Patch.Misc.IDFAArmorClass = Val(Args(1))
                            If Args(0) = "IDKFA ARMOR" Then Patch.Misc.IDKFAArmor = Val(Args(1))
                            If Args(0) = "IDKFA ARMOR CLASS" Then Patch.Misc.IDKFAArmorClass = Val(Args(1))
                            If Args(0) = "BFG CELLS/SHOT" Then Patch.Misc.BFGUsage1 = Val(Args(1))
                            If Args(0) = "MONSTERS INFIGHT" Then Patch.Misc.MonstersInfight = Val(Args(1))
                        
                        ' Extended string
                        Case strExtStrings
                            For A = 0 To Engine.ExtendedStrings.Count - 1
                                If Engine.ExtendedStrings.Keys(A) = Args(0) Then
                                    Patch.Strings(A).Text = Args(1)
                                    
                                    ' Multiline
                                    If Right$(Trim$(Args(1)), 1) = "\" Then Patch.Strings(A).Text = Mid$(Patch.Strings(A).Text, 1, Len(Patch.Strings(A).Text) - 1)
                                    
                                    While Right$(Trim$(Args(1)), 1) = "\"
                                        Line Input #openFile, Args(1)
                                        Patch.Strings(A).Text = Patch.Strings(A).Text & LTrim$(Args(1))
                                    Wend
                                    
                                    Patch.Strings(A).Text = UnEscaped(Patch.Strings(A).Text)
                                    
                                    Exit For
                                End If
                            Next A
                            
                        ' A string replacement - these should always go last in the patch file!
                        Case strString
                            
                            If Left$(UCase$(ReadLine), 4) = "TEXT" Then
                            
                                ' Get old and new lengths
                                Space1 = InStr(1, ReadLine, " ")
                                Space2 = InStr(Space1 + 1, ReadLine, " ")
                                OldLen = Val(Mid$(ReadLine, Space1 + 1, Len(ReadLine) - Space2))
                                NewLen = Val(Mid$(ReadLine, Space2 + 1))
                            
                                ' Read old and new strings
                                OldString = ReadCString(OldLen, openFile)
                                NewString = ReadCString(NewLen, openFile)
                            
                                ' Search for the replacement string
                                Replacement = False
                                For A = 0 To UBound(Engine.Strings())
                                     If Engine.Strings(A).Text = OldString Then
                                        Replacement = True
                                        Exit For
                                    End If
                                Next A
                                
                                ' Error message if no replacement could be found
                                If Replacement = False Then
                                    MsgBox "Could not find string replacement.", vbExclamation Or vbOKOnly
                                    LoadPatch = False
                                    Close #openFile
                                    Exit Function
                                End If
                                
                                ' Replace the string
                                Patch.Strings(A).length = Len(NewString)
                                Patch.Strings(A).Text = NewString
                            
                            End If
                            
                        
                    End Select

                End If

            End If


        Loop Until EOF(openFile)

    Close #openFile


    ' The loading was most likely successful
    LoadPatch = True

    Patch.filename = filename
    Patch.Saved = True
    Patch.Changed = False
    
    ' Update some "misc." table stuff. Bleargh.
    With Patch.Misc
        .BFGUsage3 = .BFGUsage2 = .BFGUsage1
        .GodHealth2 = .GodHealth1
        .MaxArmor2 = .MaxArmor1
        .MaxHealth2 = .MaxHealth1
        .MaxSoulHealth2 = .MaxSoulHealth1
    End With

    ' See if some unsupported stuff was found
    If Unsupported = True Then
        MsgBox "Unsupported entries were found in the patch file. Those entries will be ignored. This means that WhackEd3 will not save those entries again either!", vbOKOnly Or vbExclamation
    End If

End Function


' Convert a thing's bits field into usable flags (for loading BEX files)
Private Function GetThingFlags(ByVal value As String) As Long

    Dim A As Long
    Dim Result As Long
    
    
    value = Trim$(UCase$(value))
    
    ' If the bits is a value then just pass it on
    If value = "0" Or Val(value) > 0 Then
        GetThingFlags = Val(value)
        Exit Function
    End If
    
    ' It's a series of BEX flags so convert them
    Result = 0
    For A = 0 To Engine.flags.Count - 1
        If InStr(1, value, Engine.flags.Keys(A)) > 0 Then Result = SetFlag(Result, A)
    Next A
    
    ' Return the result
    GetThingFlags = Result

End Function


' Validate a patch file
Public Function Validate(ByVal filename As String) As Dictionary

    Dim openFile As Long
    Dim ReadLine As String
    
    Dim OPAssign As Long
    Dim OPComment As Long
    
    Dim Variable As String
    Dim value As String
    
    
    ' Create the new dictionary object that will hold the info
    Set Validate = New Dictionary
    
    ' Open the file
    openFile = FreeFile
    Open filename For Input As openFile
    
        Do
            
            ' Read a whole line
            Line Input #openFile, ReadLine
            
            ' Remove comments
            OPComment = InStr(1, ReadLine, "#")
            If OPComment > 0 Then
                ReadLine = Mid$(ReadLine, 1, OPComment - 1)
            End If
            
            ' Do we have an assignment?
            OPAssign = InStr(1, ReadLine, "=")
            If OPAssign > 0 Then
            
                ' Parse the variable and value
                Variable = UCase$(Trim$(Mid$(ReadLine, 1, OPAssign - 1)))
                value = Trim$(Mid$(ReadLine, OPAssign + 1))
                
                ' Store different info
                If Variable = "PATCH FORMAT" Then Validate.Add "patchversion", value
                If Variable = "DOOM VERSION" Then Validate.Add "engineversion", value
                If Variable = "ENGINE CONFIG" Then Validate.Add "engineconfig", value
                If Variable = "DATA WAD" Then Validate.Add "datawad", value
                If Variable = "IWAD" Then Validate.Add "iwad", value
            
            End If
            
        Loop Until EOF(openFile)
    
    Close openFile
    
    ' Return filename too
    Validate.Add "filename", filename

End Function


' Expand the par times array
Private Sub ExpandParList()

    ReDim Preserve Patch.Pars(Patch.nPars)
    Patch.nPars = Patch.nPars + 1

End Sub
