Attribute VB_Name = "modTableDef"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modTableDef.bas
'   Purpose: Engine\patch data table definitions
'
Option Explicit
Option Compare Binary
Option Base 0


' Tables
Public Type ThingTable              ' THING TABLE
    Id As Long                      ' Identification
    SpawnState As Long              ' Initial state
    HitPoints As Long               ' Health points
    MoveState As Long               ' Walking state
    SeePlayerSnd As Long            ' Activate\wakeup sound
    ReactionTime As Long            ' Time to react
    AttackSnd As Long               ' Attack sound
    PainState As Long               ' In pain state
    PainChance As Long              ' Pain chance
    PainSnd As Long                 ' Pain sound
    MeleeState As Long              ' Melee attack state
    AttackState As Long             ' Normal attack state
    DeathState As Long              ' Dying state
    ExplodingState As Long          ' Explosion state (missiles)
    DeathSnd As Long                ' Death sound
    Speed As Long                   ' Speed (* FRACUNIT for missiles)
    Radius As Long                  ' Radius (Total width / 2) (* FRACUNIT)
    Height As Long                  ' Height (* FRACUNIT)
    Mass As Long                    ' Mass
    MissileDamage As Long           ' Damage on missile impact
    ActSnd As Long                  ' 'Roaming' sound
    flags As Long                   ' Flags (bit settings)
    RespawnState As Long            ' Respawn state
End Type

Public Type ZDoomThingTable         ' EXTRA ZDOOM THING DATA TABLE
    Alpha As Single                 ' Alpha value
    Translucency As Long            ' Amount of translucency
    RenderMode As Byte              ' Visual render mode
End Type

Public Type StateTable              ' STATE TABLE
    Sprite As Long                  ' Sprite to display for this state
    Frame As Long                   ' Frame of the sprite to display
    Duration As Long                ' The duration in tics (1/35ths second)
    CodePointer As Long             ' Pointer to the action of this frame
    NextState As Long               ' Next state
    Parameter1 As Long              ' First parameter
    Parameter2 As Long              ' Second parameter
End Type

Public Type SoundTable              ' SOUND TABLE
    StringPointer As Long           ' Pointer to the string of this sound's name
    Singularity As Long             ' Play only one instance of this sound
    Priority As Long                ' Priority of this sound over others
    Reference As Long               ' Pointer to another sound
    Pitch As Long                   ' Pitch of referenced sound
    Volume As Long                  ' Volume of referenced sound
    Data As Long                    ' Sound data (internal)
    Usefulness As Long              ' Sound chache stuff? (internal)
    LumpNum As Long                 ' Lump number of sound lump (internal)
End Type

Public Type StringTable             ' STRING TABLE
    Offset As Long                  ' Offset of string
    Length As Long                  ' Length of string
    Text As String                  ' The string itself
End Type

Public Type WeaponTable             ' WEAPON TABLE
    AmmoType As Long                ' Type of ammo for this weapon
    DeselectState As Long           ' Deselect state
    SelectState As Long             ' Select state
    BobbingState As Long            ' Bob state
    ShootingState As Long           ' Shoot state
    MuzzleState As Long             ' Fire state
End Type

Public Type AmmoTable               ' AMMO TABLE
    Maximum As Long                 ' Maximum amount
    Pickup As Long                  ' Amount when picked up
End Type

Public Type CheatTable              ' CHEAT TABLE
    Length As Integer               ' Length of cheat string
    Text As String                  ' The cheat string itself
    Offset As Long                  ' The relative offset inside the executable
End Type

Public Type MiscTable               ' MISCELLANEOUS TABLE
    InitialHealth As Long           ' Initial player health
    InitialAmmo As Long             ' Initial amount of bullets
    MaxHealth1 As Long              ' Maxmimum amount of health
    MaxHealth2 As Long
    MaxArmor1 As Long               ' Maximum amount of armor
    MaxArmor2 As Long
    GreenArmorClass As Long         ' Green armor class number
    BlueArmorClass As Long          ' Blue armor class number
    MaxSoulHealth1 As Long          ' Maximum soulsphere health
    MaxSoulHealth2 As Long
    SoulHealth As Byte              ' Extra soulsphere health
    MegaHealth As Long              ' Extra megasphere health
    GodHealth1 As Long              ' Extra Godmode health
    GodHealth2 As Long
    IDFAArmor As Long               ' IDFA cheat extra armor
    IDFAArmorClass As Long          ' IDFA cheat armor class number
    IDKFAArmor As Long              ' IDKFA cheat extra armor
    IDKFAArmorClass As Long         ' IDKFA cheat armor class number
    BFGUsage1 As Byte               ' Amount of plasma cells used per BFG shot
    BFGUsage2 As Byte               ' (Between -128 and 127 = signed!)
    BFGUsage3 As Byte
    MonstersInfight As Byte         ' Do monsters fight each other? (202 or 221)
End Type

Public Type ParTable                ' PAR TIME TABLE
    Episode As Long                 ' Episode (0 if Doom2 MAPxy)
    Map As Long                     ' Map
    Time As Long                    ' Time in seconds
End Type

' Offsets of the misc. entries
Public Type MiscOffsetTable
    InitialHealth As Long           ' Initial health value of the player
    InitialAmmo As Long             ' Initial amount of bullets
    MaxHealth1 As Long              ' Maxmimum amount of health
    MaxHealth2 As Long
    MaxArmor1 As Long               ' Maximum amount of armor
    MaxArmor2 As Long
    GreenArmorClass As Long         ' The armor class of the green armor
    BlueArmorClass As Long          ' The armor class of the blue armor
    MaxSoulHealth1 As Long          ' The maximum health you can reach with a soulsphere
    MaxSoulHealth2 As Long
    SoulHealth As Long              ' The health you get when you grab a soulsphere
    MegaHealth As Long              ' The health you get when you grab a megasphere
    GodHealth1 As Long              ' The health you get when you enter god mode
    GodHealth2 As Long
    IDFAArmor As Long               ' The armor you recieve when you use the IDFA cheat
    IDFAArmorClass As Long          ' The armor class you get when you use the IDFA cheat
    IDKFAArmor As Long              ' The armor you recieve when you use the IDKFA cheat
    IDKFAArmorClass As Long         ' The armor class you get when you use the IDKFA cheat
    BFGUsage1 As Long               ' The amount of ammo one BFG shot uses
    BFGUsage2 As Long
    BFGUsage3 As Long
    MonstersInfight As Long         ' Will monsters engage in inter-species fights?
End Type


' A single patch
Public Type PatchData

    fileName As String
    Version As Integer
    EngineVersion As Integer

    Saved As Boolean
    Changed As Boolean


    Things() As ThingTable
    ThingNames() As String
    ZDoomThings() As ZDoomThingTable

    States() As StateTable
    CodePointers() As Long

    Sounds() As SoundTable
    SoundNames() As String * 6

    SpritePointers() As Long
    SpriteNames() As String * 4

    Strings() As StringTable
    Cheats() As CheatTable

    Weapons() As WeaponTable
    WeaponNames() As String
    
    Ammo() As AmmoTable
    AmmoNames() As String

    Misc As MiscTable

    Pars() As ParTable
    nPars As Long

End Type


' The engine data
' NOTE: There is no Par. data for the engine. It is possible, but was
'       never in the original Dehacked so its left it out for compatability's sake
Public Type EngineData

    Name As String
    Version As Integer
    ConfigFile As String
    TableFile As String
    AllowExeRead As Boolean
    ExeSize As Long
    
    Extended As Boolean
    ExtendedZDoom As Boolean


    OffsetThings As Long
    OffsetStates As Long
    OffsetSounds As Long
    OffsetStrings As Long
    OffsetWeapons As Long
    OffsetSprites As Long
    OffsetCodePointers As Long
    OffsetAmmo As Long
    OffsetCheats As Long
    OffsetMisc As Long
    OffsetDataSegment As Long

    
    Things() As ThingTable
    ThingNames() As String
    ZDoomThings() As ZDoomThingTable

    States() As StateTable
    CodePointers() As Long

    Sounds() As SoundTable
    SoundNames() As String * 6

    SpritePointers() As Long
    SpriteNames() As String * 4

    Strings() As StringTable
    Cheats() As CheatTable

    Weapons() As WeaponTable
    WeaponNames() As String
    
    Ammo() As AmmoTable
    AmmoNames() As String

    Misc As MiscTable
    MiscOffsets As MiscOffsetTable


    ' Other data - only useful for UI
    flags As Dictionary
    FlagTooltips As Dictionary


    ' Various extra stuff
    ActionInfo As Dictionary
    CodePointerMnemonics As Dictionary
    ExtendedStrings As Dictionary
    
    CodePointerToState() As Long
    CodePointerOffsets() As Long
    
End Type


' Simple struct to store sprite number and frame
Public Type SpriteInfo
    Num As Long
    Frame As Long
End Type


' Property editing forms\selectors
Public exchangeState As StateTable
Public exchangeSprite As SpriteInfo
Public exchangeSound As SoundTable
Public selectedState As Long
Public selectedAction As Long
Public stringSuccessfull As Boolean


' Data tables
Public Engine As EngineData
Public Patch As PatchData


' Fixed point math unit
Public Const FRACUNIT As Long = 65536
