Attribute VB_Name = "modTables"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modTables.bas
'   Purpose: Engine\patch table management.
'
'   Notes: Routines for loading tables from an exe are unused, but left in here
'          for possible future use.
'
Option Explicit
Option Compare Binary
Option Base 0


' Table file header
Private Const tableId As String = "ETBL"
Private Const tableVersionHigh As Byte = "1"
Private Const tableVersionLow As Byte = "3"

' DLL declares
Private Declare Sub CheatToString Lib "whacked3" (ByVal CheatData As String, ByVal OutData As String, ByVal Length As Long)


' Load the table data
Public Function loadEngineTables() As Boolean

    ' No table file to read from
    If Not FS.FileExists(APP_PATH & "tables\" & Engine.TableFile) Then
        MsgBox "No table file for engine configuration.", vbExclamation Or vbOKOnly
        Exit Function
    End If
        
    ' Load data from table file
    If modTables.LoadTable(APP_PATH & "tables\" & Engine.TableFile) = False Then Exit Function

    loadEngineTables = True

End Function


' Load a table from an exe file
Private Sub loadExeTableFile(ByVal fileName As String)

    Dim A As Long
    Dim openFile As Long
    Dim ReadChar As String * 1
    Dim NewString As String


    appLog.Add "  Loading tables from " & fileName & "..."

    ' Check if the file exists
    If Not FS.FileExists(fileName) Then
        MsgBox "Engine executable file does not exists.", vbExclamation Or vbOKOnly
        Exit Sub
    End If
        
    
    ' Check the executable's validacy
    If FileLen(fileName) <> Engine.ExeSize Then
        MsgBox "Engine executable filesize is invalid.", vbExclamation Or vbOKOnly
        Exit Sub
    End If


    openFile = FreeFile
    Open fileName For Binary As #openFile

        ' Things
        Seek #openFile, Engine.OffsetThings
        Get #openFile, , Engine.Things


        ' States
        Seek #openFile, Engine.OffsetStates
        Get #openFile, , Engine.States


        ' Sprites
        Seek #openFile, Engine.OffsetSprites
        Get #openFile, , Engine.SpritePointers

        For A = 0 To UBound(Engine.SpritePointers())
            Seek #openFile, Engine.OffsetDataSegment + Engine.SpritePointers(A)
            Get #openFile, , Engine.SpriteNames(A)
        Next A


        ' Sounds
        Seek #openFile, Engine.OffsetSounds
        Get #openFile, , Engine.Sounds

        For A = 0 To UBound(Engine.Sounds())
            Seek #openFile, Engine.OffsetDataSegment + Engine.Sounds(A).StringPointer
            Get #openFile, , Engine.SoundNames(A)
            Engine.SoundNames(A) = UCase(TrimNull(Engine.SoundNames(A)))
        Next A


        ' Codepointers (actions)
        For A = 0 To UBound(Engine.CodePointerOffsets())
            Seek #openFile, Engine.OffsetCodePointers + Engine.CodePointerOffsets(A)
            Get #openFile, , Engine.CodePointers(Engine.CodePointerToState(A))
        Next A

        
        ' Strings
        Seek #openFile, Engine.OffsetStrings
        
        For A = 0 To UBound(Engine.Strings())
            
            ' Read until the next 4-byte boundary
            If (Loc(openFile) Mod 4) <> 0 Then
                Do
                    Get #openFile, , ReadChar
                Loop Until (Loc(openFile) Mod 4) = 0
            End If
            
            ' Store offset
            Engine.Strings(A).Offset = Loc(openFile)
            
            ' Read the text itself, one char at a time
            ReadChar = vbNullString
            Do
              
                ' Read a single character
                Get #openFile, , ReadChar
              
                ' Exit the loop when a null char is found
                If Asc(ReadChar) = 0 Then Exit Do
              
                ' Add it to the entire string
                Engine.Strings(A).Text = Engine.Strings(A).Text & ReadChar
                      
            Loop
            
            Engine.Strings(A).Length = Len(Engine.Strings(A).Text)
            
        Next A
        
        
        ' Weapons
        Seek #openFile, Engine.OffsetWeapons
        Get #openFile, , Engine.Weapons
        
        
        ' Ammo
        ' NOTE: The last two items aren't loaded, they do not
        '       have an actual entry in the ammo table.
        Seek #openFile, Engine.OffsetAmmo
        
        For A = 0 To UBound(Engine.AmmoNames()) - 2
            Get #openFile, , Engine.Ammo(A).Maximum
        Next A
        For A = 0 To UBound(Engine.AmmoNames()) - 2
            Get #openFile, , Engine.Ammo(A).Pickup
        Next A
        
        
        ' Cheats
        For A = 0 To UBound(Engine.Cheats())
            
            ' Seek
            Seek #openFile, Engine.OffsetCheats + Engine.Cheats(A).Offset
            
            ' Read the string
            Engine.Cheats(A).Text = Space$(Engine.Cheats(A).Length)
            NewString = Space$(Engine.Cheats(A).Length)
            Get #openFile, , Engine.Cheats(A).Text
            
            ' Decrypt it
            CheatToString Engine.Cheats(A).Text, NewString, Engine.Cheats(A).Length
            Engine.Cheats(A).Text = NewString
            
        Next A
        
        
        ' Misc stuff (ewwwww)
        Seek #openFile, Engine.MiscOffsets.BFGUsage1
        Get #openFile, , Engine.Misc.BFGUsage1
        Seek #openFile, Engine.MiscOffsets.BFGUsage2
        Get #openFile, , Engine.Misc.BFGUsage2
        Seek #openFile, Engine.MiscOffsets.BFGUsage3
        Get #openFile, , Engine.Misc.BFGUsage3
        Seek #openFile, Engine.MiscOffsets.BlueArmorClass
        Get #openFile, , Engine.Misc.BlueArmorClass
        Seek #openFile, Engine.MiscOffsets.GodHealth1
        Get #openFile, , Engine.Misc.GodHealth1
        Seek #openFile, Engine.MiscOffsets.GodHealth2
        Get #openFile, , Engine.Misc.GodHealth2
        Seek #openFile, Engine.MiscOffsets.GreenArmorClass
        Get #openFile, , Engine.Misc.GreenArmorClass
        Seek #openFile, Engine.MiscOffsets.IDFAArmor
        Get #openFile, , Engine.Misc.IDFAArmor
        Seek #openFile, Engine.MiscOffsets.IDFAArmorClass
        Get #openFile, , Engine.Misc.IDFAArmorClass
        Seek #openFile, Engine.MiscOffsets.IDKFAArmor
        Get #openFile, , Engine.Misc.IDKFAArmor
        Seek #openFile, Engine.MiscOffsets.IDKFAArmorClass
        Get #openFile, , Engine.Misc.IDKFAArmorClass
        Seek #openFile, Engine.MiscOffsets.InitialAmmo
        Get #openFile, , Engine.Misc.InitialAmmo
        Seek #openFile, Engine.MiscOffsets.InitialHealth
        Get #openFile, , Engine.Misc.InitialHealth
        Seek #openFile, Engine.MiscOffsets.MaxArmor1
        Get #openFile, , Engine.Misc.MaxArmor1
        Seek #openFile, Engine.MiscOffsets.MaxArmor2
        Get #openFile, , Engine.Misc.MaxArmor2
        Seek #openFile, Engine.MiscOffsets.MaxHealth1
        Get #openFile, , Engine.Misc.MaxHealth1
        Seek #openFile, Engine.MiscOffsets.MaxHealth2
        Get #openFile, , Engine.Misc.MaxHealth2
        Seek #openFile, Engine.MiscOffsets.MaxSoulHealth1
        Get #openFile, , Engine.Misc.MaxSoulHealth1
        Seek #openFile, Engine.MiscOffsets.MaxSoulHealth2
        Get #openFile, , Engine.Misc.MaxSoulHealth2
        Seek #openFile, Engine.MiscOffsets.MegaHealth
        Get #openFile, , Engine.Misc.MegaHealth
        Seek #openFile, Engine.MiscOffsets.MonstersInfight
        Get #openFile, , Engine.Misc.MonstersInfight
        Seek #openFile, Engine.MiscOffsets.SoulHealth
        Get #openFile, , Engine.Misc.SoulHealth



    Close #openFile
    modTables.DisplayTableInfo

    ' Copy engine data over to patch data
    modTables.ResetPatch

End Sub


' Return the name of a state number (sprite name + sprite frame)
Public Function StateNameFromNum(ByVal StateNum As Long) As String

    If StateNum = 0 Then
        StateNameFromNum = "-"
        Exit Function
    End If

    If GetFlag(Patch.States(StateNum).Frame, 15) Then
        StateNameFromNum = Patch.SpriteNames(Patch.States(StateNum).Sprite) & Chr$(Patch.States(StateNum).Frame - (2 ^ 15) + 65)
    Else
        StateNameFromNum = Patch.SpriteNames(Patch.States(StateNum).Sprite) & Chr$(Patch.States(StateNum).Frame + 65)
    End If

End Function


' Return the name of a state reference (sprite name + sprite frame)
Public Function StateNameFromRef(ByRef State As StateTable) As String

    If GetFlag(State.Frame, 15) Then
        StateNameFromRef = Patch.SpriteNames(State.Sprite) & Chr$(State.Frame - (2 ^ 15) + 65)
    Else
        StateNameFromRef = Patch.SpriteNames(State.Sprite) & Chr$(State.Frame + 65)
    End If

End Function


' Return the name of a sound
Public Function SoundName(ByVal SoundNum As Long) As String

    If SoundNum = 0 Then
        SoundName = "-"
        Exit Function
    End If

    SoundName = Patch.SoundNames(SoundNum)

End Function


' Copy the engine table stuff over to the patch table
Public Sub ResetPatch()

    Dim A As Long
    
    
    ' Tables
    Patch.Ammo = Engine.Ammo
    Patch.AmmoNames = Engine.AmmoNames
    Patch.Cheats = Engine.Cheats
    Patch.CodePointers = Engine.CodePointers
    Patch.Misc = Engine.Misc
    Patch.SoundNames = Engine.SoundNames
    Patch.Sounds = Engine.Sounds
    Patch.SpriteNames = Engine.SpriteNames
    Patch.SpritePointers = Engine.SpritePointers
    Patch.States = Engine.States
    Patch.Strings = Engine.Strings
    Patch.ThingNames = Engine.ThingNames
    Patch.Things = Engine.Things
    Patch.ZDoomThings = Engine.ZDoomThings
    Patch.WeaponNames = Engine.WeaponNames
    Patch.Weapons = Engine.Weapons
        
    ' Misc data
    Patch.EngineVersion = Engine.Version
    
    ' Reset par data
    Patch.nPars = 0
    ReDim Patch.Pars(0)
    
    ' Exntded mode filename extension
    If Engine.Extended = True Then
        Patch.fileName = getSpecialPath(myDocuments) & "\unnamed.bex"
    Else
        Patch.fileName = getSpecialPath(myDocuments) & "\unnamed.deh"
    End If
    
    ' Patch file info
    Patch.Version = 6
    Patch.Changed = False
    Patch.Saved = False
    
    ' Transfer strings if in Extended mode
    If Engine.Extended = True Then
    
        For A = 0 To UBound(Patch.Strings())
            Patch.Strings(A).Text = UnEscaped(Engine.ExtendedStrings.Items(A))
        Next A
    
    End If

    appLog.Add "  Reset patch data"

End Sub


' Save patch data to a table file
Public Sub SaveTable(ByVal fileName As String)

    Dim A As Long
    Dim openFile As Long


    ' Delete old table file if it exists
    If Not FS.FileExists(fileName) <> vbNullString Then FS.DeleteFile fileName

    ' Open the table file
    openFile = FreeFile
    Open fileName For Binary As #openFile

        ' Write header
        Put #openFile, , tableId
        Put #openFile, , tableVersionHigh
        Put #openFile, , tableVersionLow


        ' Write things
        Put #openFile, , CLng(UBound(Patch.Things()))
        Put #openFile, , Patch.ThingNames
        Put #openFile, , Patch.Things
        Put #openFile, , Patch.ZDoomThings

        ' Write states
        Put #openFile, , CLng(UBound(Patch.States()))
        Put #openFile, , Patch.States

        ' Write codepointers (actions)
        Put #openFile, , CLng(UBound(Patch.CodePointers()))
        Put #openFile, , Patch.CodePointers

        ' Write sounds
        Put #openFile, , CLng(UBound(Patch.Sounds()))
        Put #openFile, , Patch.SoundNames
        Put #openFile, , Patch.Sounds

        ' Write sprites
        Put #openFile, , CLng(UBound(Patch.SpritePointers()))
        Put #openFile, , Patch.SpritePointers
        Put #openFile, , Patch.SpriteNames
        
        ' Write strings
        If Engine.Extended = False Then
            Put #openFile, , CLng(UBound(Patch.Strings()))
            
            For A = 0 To UBound(Patch.Strings())
                Put #openFile, , Patch.Strings(A).Length
                Put #openFile, , Patch.Strings(A).Offset
                Put #openFile, , Patch.Strings(A).Text
            Next A
        End If
        
        ' Write weapons
        Put #openFile, , CLng(UBound(Patch.WeaponNames()))
        Put #openFile, , Patch.WeaponNames
        Put #openFile, , Patch.Weapons
        
        ' Write ammo
        Put #openFile, , CLng(UBound(Patch.Ammo))
        Put #openFile, , Patch.AmmoNames
        Put #openFile, , Patch.Ammo
        
        ' Write cheats
        Put #openFile, , CLng(UBound(Patch.Cheats))
        
        For A = 0 To UBound(Patch.Cheats())
            Put #openFile, , CLng(Len(Patch.Cheats(A).Text))
            Put #openFile, , Patch.Cheats(A)
        Next A
        
        ' Write misc stuff
        Put #openFile, , Patch.Misc
        

    Close #openFile

End Sub


' Load patch data from a table file
Public Function LoadTable(ByVal fileName As String) As Boolean

    Dim A As Long
    Dim openFile As Long
    Dim ArrayBound As Long
    Dim StringLen As Long
    Dim cId As String * 4
    Dim cVersionHigh As Byte
    Dim cVersionLow As Byte


    LoadTable = False
    appLog.Add "  Loading tables from " & fileName & "..."

    ' Open the table file
    openFile = FreeFile
    Open fileName For Binary As #openFile

        ' Read header
        Get #openFile, , cId
        Get #openFile, , cVersionHigh
        Get #openFile, , cVersionLow


        ' Check if the file is valid
        If cId <> tableId Then
            MsgBox "Invalid table Id.", vbExclamation Or vbOKOnly
            Exit Function
        End If

        If cVersionHigh <> tableVersionHigh Or cVersionLow <> tableVersionLow Then
            MsgBox "Invalid table version.", vbExclamation Or vbOKOnly
            Exit Function
        End If


        ' Read things
        Get #openFile, , ArrayBound
        ReDim Engine.ThingNames(ArrayBound)
        ReDim Engine.Things(ArrayBound)
        ReDim Engine.ZDoomThings(ArrayBound)

        Get #openFile, , Engine.ThingNames
        Get #openFile, , Engine.Things
        Get #openFile, , Engine.ZDoomThings


        ' Read states
        Get #openFile, , ArrayBound
        ReDim Engine.States(ArrayBound)

        Get #openFile, , Engine.States


        ' Read codepointer data
        Get #openFile, , ArrayBound
        ReDim Engine.CodePointers(ArrayBound)

        Get #openFile, , Engine.CodePointers


        ' Read sounds
        Get #openFile, , ArrayBound
        ReDim Engine.SoundNames(ArrayBound)
        ReDim Engine.Sounds(ArrayBound)

        Get #openFile, , Engine.SoundNames
        Get #openFile, , Engine.Sounds


        ' Read sprites
        Get #openFile, , ArrayBound
        ReDim Engine.SpritePointers(ArrayBound)
        ReDim Engine.SpriteNames(ArrayBound)

        Get #openFile, , Engine.SpritePointers
        Get #openFile, , Engine.SpriteNames
        
        
        ' Do not read strings if in Extended mode
        If Engine.Extended = False Then
            Get #openFile, , ArrayBound
            ReDim Engine.Strings(ArrayBound)
            
            For A = 0 To UBound(Engine.Strings())
                Get #openFile, , Engine.Strings(A).Length
                Get #openFile, , Engine.Strings(A).Offset
                
                Engine.Strings(A).Text = Space$(Engine.Strings(A).Length)
                Get #openFile, , Engine.Strings(A).Text
            Next A
        End If
            
        
        ' Read weapons
        Get #openFile, , ArrayBound
        ReDim Engine.Weapons(ArrayBound)
        ReDim Engine.WeaponNames(ArrayBound)
        
        Get #openFile, , Engine.WeaponNames
        Get #openFile, , Engine.Weapons
        
        
        ' Read ammo
        Get #openFile, , ArrayBound
        ReDim Engine.Ammo(ArrayBound)
        ReDim Engine.AmmoNames(ArrayBound)
        
        Get #openFile, , Engine.AmmoNames
        Get #openFile, , Engine.Ammo
        
        
        ' Read cheats
        Get #openFile, , ArrayBound
        ReDim Engine.Cheats(ArrayBound)
        
        For A = 0 To ArrayBound
            Get #openFile, , StringLen
            
            Engine.Cheats(A).Text = Space$(StringLen)
            Get #openFile, , Engine.Cheats(A)
        Next A
        
        
        ' Read misc data
        Get #openFile, , Engine.Misc
        

    Close #openFile

    modTables.DisplayTableInfo

    ' Copy engine data over to patch data
    modTables.ResetPatch
    
    LoadTable = True

End Function


' Display information about the currently loaded table
Public Sub DisplayTableInfo()

    appLog.Add "    Loaded " & UBound(Engine.Things()) + 1 & " things"
    appLog.Add "    Loaded " & UBound(Engine.States()) + 1 & " states"
    appLog.Add "    Loaded " & UBound(Engine.Sounds()) + 1 & " sounds"
    appLog.Add "    Loaded " & UBound(Engine.SpritePointers()) + 1 & " sprites"
    appLog.Add "    Loaded " & Engine.CodePointerMnemonics.Count & " code pointers"
    appLog.Add "    Loaded " & UBound(Engine.Strings()) + 1 & " strings"
    appLog.Add "    Loaded " & UBound(Engine.WeaponNames()) + 1 & " weapons"
    appLog.Add "    Loaded " & UBound(Engine.Ammo()) + 1 & " ammo entries"
    appLog.Add "    Loaded " & UBound(Engine.Cheats()) + 1 & " cheat strings"

End Sub
