Attribute VB_Name = "modUtil"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modUtil.bas
'   Purpose: Miscellaneous utility functions, for WhackEd3 only.
'
Option Explicit
Option Compare Binary
Option Base 0


' Enum values for special paths for GetSpecialpath
Public Enum specialPathEnum
    myDocuments = 1
    applicationData = 2
    Temp = 3
    userDesktop = 4
End Enum


' Logical font structure
Private Type LOGFONT
    lfHeight As Long
    lfWidth As Long
    lfEscapement As Long
    lfOrientation As Long
    lfWeight As Long
    lfItalic As Byte
    lfUnderline As Byte
    lfStrikeOut As Byte
    lfCharSet As Byte
    lfOutPrecision As Byte
    lfClipPrecision As Byte
    lfQuality As Byte
    lfPitchAndFamily As Byte
    lfFaceName(31) As Byte
End Type

' Common controls init data
Private Type tagInitCommonControlsEx
   lngSize As Long
   lngICC As Long
End Type


Private Type WAVEFORMATEX
    wFormatTag As Integer
    nChannels As Integer
    nSamplesPerSec As Long
    nAvgBytesPerSec As Long
    nBlockAlign As Integer
    wBitsPerSample As Integer
End Type

Private Type RIFFCHUNK
    id As String * 4
    Size As Long
End Type


' Sound API constants
Private Const SND_SYNC = &H0            ' Play synchronously (default)
Private Const SND_ASYNC = &H1           ' Play asynchronously
Private Const SND_NODEFAULT = &H2       ' Silence not default, if sound not found
Private Const SND_MEMORY = &H4          ' lpszSoundName points to a memory file
Private Const SND_LOOP = &H8            ' Loop the sound until next sndPlaySound
Private Const SND_NOSTOP = &H10         ' Don't stop any currently playing sound
Private Const SND_PURGE = &H40          ' Purge non-static events for task
Private Const SND_APPLICATION = &H80    ' Look for application specific association
Private Const SND_NOWAIT = &H2000       ' Don't wait if the driver is busy
Private Const SND_ALIAS = &H10000       ' Name is a WIN.INI [sounds] entry
Private Const SND_FILENAME = &H20000    ' Name is a file name
Private Const SND_RESOURCE = &H40004    ' Name is a resource name or atom
Private Const SND_ALIAS_ID = &H110000   ' Name is a WIN.INI [sounds] entry identifier

' To get default system font with SystemParameterInfo
Private Const SPI_GETICONTITLELOGFONT As Long = 31

' Image constant
' TODO: Make it dependant on IWAD used
Public Const ImageTransparency As Byte = 0

' SendMessage
Private Const HWND_BROADCAST As Long = &HFFFF
Private Const WM_FONTCHANGE As Long = &H1D

' Special folder constants
Private Const CSIDL_APPDATA As Long = &H1A
Private Const CSIDL_MYDOCUMENTS As Long = &H5
Private Const CSIDL_DESKTOP As Long = &H0
Private Const CSIDL_FLAG_CREATE As Long = &H8000

' Maximum length of some path variables
Private Const MAX_PATH = 260


' Sound playing
Private SoundPlayStart As Long
Private SoundLength As Long

' API calls
Private Declare Function playSound Lib "winmm" Alias "PlaySoundA" (ByVal lpszName As String, ByVal hModule As Long, ByVal dwFlags As Long) As Long
Private Declare Function GetTickCount Lib "kernel32" () As Long
Private Declare Function SystemParametersInfo Lib "user32" Alias "SystemParametersInfoA" (ByVal uiAction As Long, ByVal uiParam As Long, ByVal ptrParam As Long, ByVal fWinIni As Long) As Long
Private Declare Function CreateDC Lib "gdi32" Alias "CreateDCA" (ByVal ptrDriver As String, ByVal ptrDevice As Long, ByVal ptrOutput As Long, ByVal printerData As Long) As Long
Private Declare Function StretchBlt Lib "gdi32" (ByVal hDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal nSrcWidth As Long, ByVal nSrcHeight As Long, ByVal dwRop As Long) As Long
Private Declare Function AddFontResource Lib "gdi32" Alias "AddFontResourceA" (ByVal filename As String) As Long
Private Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lparam As Long) As Long
Private Declare Function SHGetFolderPath Lib "shfolder" Alias "SHGetFolderPathA" (ByVal hwndOwner As Long, ByVal nFolder As Long, ByVal hToken As Long, ByVal dwFlags As Long, ByVal lpszPath As String) As Long
Private Declare Function GetTempPath Lib "kernel32" Alias "GetTempPathA" (ByVal nBufferLength As Long, ByVal lpBuffer As String) As Long
Private Declare Function InitCommonControlsEx Lib "comctl32" (iCCEx As tagInitCommonControlsEx) As Boolean

Public Declare Sub CopyMemory Lib "kernel32.dll" Alias "RtlMoveMemory" (ByRef dest As Any, ByRef Source As Any, ByVal bytes As Long)
Public Declare Sub FillMemory Lib "kernel32.dll" Alias "RtlFillMemory" (Destination As Any, ByVal length As Long, ByVal Fill As Byte)


' WhackEd3 DLL functions
Public Declare Function SetFlag Lib "whacked3" (ByVal SetVal As Long, ByVal FlagNum As Long) As Long
Public Declare Function RemoveFlag Lib "whacked3" (ByVal SetVal As Long, ByVal FlagNum As Long) As Long
Public Declare Function GetFlag Lib "whacked3" (ByVal UseVal As Long, ByVal FlagNum As Long) As Long


' Validate a value in a textbox and return it
Public Function validateValue(ByRef Control As TextBox, ByVal minVal As Long, ByVal maxVal As Long) As Long

    Dim Value As Long
    
    
    Value = Val(Control.Text)
    
    If Value < minVal Then Value = minVal
    If Value > maxVal Then Value = maxVal
    
    If Value <> Val(Control.Text) Then Control.Text = Value
    validateValue = Value

End Function


' Render a sprite + frame to a picturebox, choosing the right rotation
Public Function previewSprite(ByRef picBox As PictureBox, ByVal spriteNum As Long, ByVal frameNum As Long) As String

    Dim Lump0 As Dictionary
    Dim Lump1 As Dictionary
    Dim useLump As Dictionary
    Dim WAD As clsWad
    

    ' Search for both sprite lumps with and without rotations
    Set Lump0 = modWADList.WADNumForNameSprite(Engine.SpriteNames(spriteNum), Chr$(frameNum + 65) & "0")
    Set Lump1 = modWADList.WADNumForNameSprite(Engine.SpriteNames(spriteNum), Chr$(frameNum + 65) & "1")
    
    ' Determine lump to use
    ' This way sprites that have no rotations in the IWAD, but have in a PWAD, will be taken from the PWAD instead.
    If Lump0 Is Nothing Then
        Set useLump = Lump1
    ElseIf Lump1 Is Nothing Then
        Set useLump = Lump0
    ElseIf Lump0("wadIndex") > Lump1("wadIndex") Then
        Set useLump = Lump0
    ElseIf Lump1("wadIndex") > Lump0("wadIndex") Then
        Set useLump = Lump0
    End If
    
    ' Render it
    If Not (useLump Is Nothing) Then
        Set WAD = useLump("class")
        modUtil.renderDoomImage WAD.ReadLumpString(useLump("num")), paletteData, picBox, 1
        previewSprite = useLump("name")
        
    Else
        picBox.Cls
        previewSprite = vbNullString
    
    End If

End Function


' Auto-complete a combobox
Public Function autoComplete(ByRef Box As ComboBox) As Long

    Dim A As Long
    Dim boxPart As String
    Dim curLen As Long
    
    
    curLen = Len(Box.Text)
    boxPart = Left$(LCase$(Box.Text), curLen)
    
    For A = 0 To Box.ListCount - 1
        If Left(LCase$(Box.List(A)), curLen) = boxPart Then
            autoComplete = A
            Exit Function
        End If
    Next A
    
    autoComplete = -1

End Function


' Select the Text contents of a control
Public Sub selectAll(ByRef Control As Control)

    Control.SelStart = 0
    Control.SelLength = Len(Control.Text)

End Sub


' Install a font temporarily
Public Sub installFont(ByVal filename As String)

    ' Add font as resource (removed when rebooted)
    If AddFontResource(filename) = 0 Then appLog.Add "Unable to install font " & filename
    
    ' Broadcast to other windows that new fonts have been installed
    SendMessage HWND_BROADCAST, WM_FONTCHANGE, 0, 0

End Sub


' Render a single Doom image to a picture box
Public Sub renderDoomImage(ByRef data As String, ByRef palData As String, ByRef picBox As PictureBox, ByVal iScale As Long)

    Dim Bitmap As New clsBitmap
    Dim Width As Long, Height As Long
    Dim X As Long, Y As Long
    
    
    picBox.Cls
    
    Width = CVI(Mid$(data, 1, 2))
    Height = CVI(Mid$(data, 3, 2))
    
    X = picBox.Width / 2 - (Width / 2)
    Y = picBox.Height / 2 - (Height / 2)
    
    Bitmap.setPalette palData
    Bitmap.Create Width, Height, ImageTransparency
    Bitmap.renderDoomImage data, 0, 0
    
    StretchBlt picBox.hDC, X, Y, Width * iScale, Height * iScale, Bitmap.hDC, 0, 0, Width, Height, vbSrcCopy
    picBox.Refresh

End Sub


' Get the default system font and apply it to all controls in this program
Public Function applyDefaultSystemFont() As Boolean

    Dim useFont As LOGFONT
    Dim myForm As Form
    Dim myControl As Control
    Dim desktopDC As Long
    Dim oldFont As StdFont

    
    ' Get default system font
    If SystemParametersInfo(SPI_GETICONTITLELOGFONT, LenB(useFont), VarPtr(useFont), 0) = 0 Then Exit Function
    
    ' Get desktop device context
    desktopDC = CreateDC("DISPLAY", 0, 0, 0)
    If desktopDC = 0 Then Exit Function
    
    
    ' Apply font object to all controls on all loaded forms
    For Each myForm In Forms
        For Each myControl In myForm.Controls
            
            ' Only apply font to objects that need it
            Select Case TypeName(myControl)
                Case "CommandButton", "Label", "TextBox", "ComboBox", "CheckBox", "ctlFancyList", "OptionButton", "PictureBox", "ListBox", "DTPicker"
                    
                    If myControl.Tag <> 2 Then
                        
                        Set oldFont = myControl.Font
                    
                        ' Create new font object for each control to keep specific size and styles
                        Set myControl.Font = New StdFont
                        With myControl.Font
                            .name = StrConv(useFont.lfFaceName, vbUnicode)
                            .Charset = useFont.lfCharSet
                            .Bold = oldFont.Bold
                            .Italic = oldFont.Italic
                            .Underline = oldFont.Underline
                            .Size = oldFont.Size
                        End With
                        
                    End If

            End Select
            
        Next myControl
    Next myForm
    
    
    applyDefaultSystemFont = True

End Function


' Trim (literally remove) NULL characters from a string
Public Function TrimNull(ByVal inString As String) As String

    TrimNull = Trim$(Replace(inString, Chr$(0), " "))

End Function


' Write a state into a table (boundary checks)
Public Sub WriteStateUI(ByRef StateVar As Long, ByVal StateVal As Long, ByRef txtStateNum As TextBox)

    If StateVal < 0 Then
        StateVal = 0
        txtStateNum.Text = 0
    End If
    If StateVal > UBound(Patch.States()) Then
        StateVal = UBound(Patch.States())
        txtStateNum.Text = UBound(Patch.States())
    End If

    StateVar = StateVal
    patchHasChanged

End Sub


' Select and display a single sound in the Sound editor
Public Sub SelectSingleSound(ByRef SoundNum As Long)

    With frmEditSounds.lvwSounds
        .MultiSelect = False
        .ListItems(SoundNum).Selected = True
        .ListItems(SoundNum).EnsureVisible
        .MultiSelect = True
    End With

End Sub


' Write a sound into a table (boundary checks)
Public Sub WriteSoundUI(ByRef SoundVar, ByVal SoundVal, ByRef txtSoundNum As TextBox)

    If SoundVal < 0 Then
        SoundVal = 0
        txtSoundNum.Text = 0
    End If
    If SoundVal > UBound(Patch.Sounds()) Then
        SoundVal = UBound(Patch.Sounds())
        txtSoundNum.Text = UBound(Patch.Sounds())
    End If

    SoundVar = SoundVal
    patchHasChanged

End Sub


' Update a single entry in a state listview
Public Sub UpdateStateEntry(ByVal listIndex As Long, ByRef StateNum As Long, ByRef lvwStates As ListView)

    With lvwStates.ListItems(listIndex)
        .Text = StateNum
        .ListSubItems(1).Text = StateNameFromNum(StateNum)
        .ListSubItems(2).Text = Patch.States(StateNum).Sprite
    
        ' Lit
        If Patch.States(StateNum).Frame And (2 ^ 15) Then
            .ListSubItems(3).Text = Patch.States(StateNum).Frame And Not (2 ^ 15)
            .ListSubItems(4).Text = "Yes"
        Else
            .ListSubItems(3).Text = Patch.States(StateNum).Frame
            .ListSubItems(4).Text = "No"
        End If
    
        .ListSubItems(5).Text = Patch.States(StateNum).NextState
        .ListSubItems(6).Text = Patch.States(StateNum).Duration
        .ListSubItems(7).Text = Patch.States(StateNum).Parameter1
        .ListSubItems(8).Text = Patch.States(StateNum).Parameter2
        .ListSubItems(9).Text = Engine.CodePointerMnemonics.Item(Trim$(Str$(Patch.CodePointers(StateNum))))
        
        .ForeColor = 0
        .ListSubItems(1).ForeColor = 0
        .ListSubItems(4).ForeColor = 0
        .ListSubItems(9).ForeColor = 0
    End With

End Sub


' Preview a sound
Public Sub previewSound(ByVal SoundNum As Long)

    Dim soundData() As Byte
    Dim Lump As Dictionary
    Dim sampleRate As Integer
    Dim soundLen As Long
    Dim Channel As Long
    
    Static Sample As Long
    

    ' Can't play reference sounds
    If Patch.Sounds(SoundNum).Reference <> 0 Then Exit Sub

    ' We can't play the dummy sound either...
    If SoundNum = 0 Then Exit Sub

    ' Free up possible previous sample so we don't keep consuming more memory
    ' Also stops any previously playing sample
    BASS_SampleFree Sample
    
    Set Lump = modWADList.WADNumForName("DS" & TrimNull(SoundName(SoundNum)))
    If Lump Is Nothing Then Exit Sub
    
    ' Read data and get info
    soundData = Lump("class").ReadLump(Lump("num"))
    soundLen = UBound(soundData) - 8
    CopyMemory ByVal VarPtr(sampleRate), ByVal VarPtr(soundData(2)), 2
    
    ' Create sample and feed it data
    Sample = BASS_SampleCreate(soundLen, sampleRate, 1, 1, BASS_SAMPLE_8BITS)
    If Sample = 0 Then
        appLog.Add "Could not create sample. BASS error code " & BASS_ErrorGetCode
        Exit Sub
    End If
    BASS_SampleSetData Sample, ByVal VarPtr(soundData(8))
    
    ' Get channel for sample
    Channel = BASS_SampleGetChannel(Sample, 1)
    If Channel = 0 Then
        appLog.Add "Could not create channel from sample. BASS error code " & BASS_ErrorGetCode
        Exit Sub
    End If
    
    ' Play
    appLog.Add "Playing sound lump " & Lump("num") & " (" & Lump("name") & ")"
    appLog.Add sampleRate & " Hz, " & soundLen & " bytes"
    
    BASS_ChannelPlay Channel, 1
        
End Sub


' Call this when the patch structure is changed
Public Sub patchHasChanged()

    Patch.Changed = True
    updateSavingFunctions

End Sub


' Rework the saving function's availability
Public Sub updateSavingFunctions()

    If Patch.Changed = True Then
        frmMain.tlbProgram.Buttons(4).Enabled = True
        frmMain.FileSave.Enabled = True
    Else
        frmMain.tlbProgram.Buttons(4).Enabled = False
        frmMain.FileSave.Enabled = False
    End If

End Sub


' Update a single entry in a sound listview
Public Sub UpdateSoundEntry(ByRef SoundNum As Long, ByRef lvwSounds As ListView)

    lvwSounds.ListItems(SoundNum + 1).Text = SoundNum
    lvwSounds.ListItems(SoundNum + 1).ListSubItems(1).Text = SoundName(SoundNum)
    lvwSounds.ListItems(SoundNum + 1).ListSubItems(2).Text = Patch.Sounds(SoundNum).StringPointer

    If Patch.Sounds(SoundNum).Singularity = 1 Then
        lvwSounds.ListItems(SoundNum + 1).ListSubItems(3).Text = "Yes"
    Else
        lvwSounds.ListItems(SoundNum + 1).ListSubItems(3).Text = "No"
    End If

    lvwSounds.ListItems(SoundNum + 1).ListSubItems(4).Text = Patch.Sounds(SoundNum).priority
    lvwSounds.ListItems(SoundNum + 1).ListSubItems(5).Text = Patch.Sounds(SoundNum).Reference
    lvwSounds.ListItems(SoundNum + 1).ListSubItems(6).Text = Patch.Sounds(SoundNum).Pitch
    lvwSounds.ListItems(SoundNum + 1).ListSubItems(7).Text = Patch.Sounds(SoundNum).volume

End Sub


' Update a single entry in a string listview
Public Sub UpdateStringEntry(ByVal StringNum As Long, ByRef lvwStrings As ListView)

    lvwStrings.ListItems(StringNum + 1).Text = StringNum
    lvwStrings.ListItems(StringNum + 1).ListSubItems(2).Text = Escaped(Patch.Strings(StringNum).Text)
        
    ' Boom extension has Mnemonic instead of Length
    If Engine.Extended = True Then
        lvwStrings.ListItems(StringNum + 1).ListSubItems(1).Text = Engine.ExtendedStrings.Keys(StringNum)
    Else
        lvwStrings.ListItems(StringNum + 1).ListSubItems(1).Text = Patch.Strings(StringNum).length
    End If

End Sub


' Strip leading and trailing quotes from a string
Public Function StripQuotes(ByRef QuoteString As String) As String

    QuoteString = Trim$(QuoteString)

    If Left$(QuoteString, 1) = Chr$(34) Then Mid$(QuoteString, 1, 1) = " "
    If Right$(QuoteString, 1) = Chr$(34) Then Mid$(QuoteString, Len(QuoteString), 1) = " "

    StripQuotes = Trim$(QuoteString)

End Function


' See clsConfiguration.Escaped
' Used for the String Editor\Property form
Public Function Escaped(ByVal Value As String) As String

    Escaped = Value

    Escaped = Replace$(Escaped, "\", "\\")
    Escaped = Replace$(Escaped, vbLf, "\n")
    Escaped = Replace$(Escaped, vbCr, "\r")
    Escaped = Replace$(Escaped, vbTab, "\t")
    Escaped = Replace$(Escaped, """", "\""")
     
End Function

Public Function UnEscaped(ByVal Value As String) As String

    UnEscaped = Value
    
    UnEscaped = Replace$(UnEscaped, "\\", "\")
    UnEscaped = Replace$(UnEscaped, "\n", vbLf)
    UnEscaped = Replace$(UnEscaped, "\r", vbCr)
    UnEscaped = Replace$(UnEscaped, "\t", vbTab)
    UnEscaped = Replace$(UnEscaped, "\""", """")

End Function

Public Function Returned(ByVal Value As String) As String

    Returned = Value
    
    Returned = Replace$(Returned, "\n", "\n" & vbCrLf)

End Function

Public Function UnReturned(ByVal Value As String) As String

    UnReturned = Value
    
    UnReturned = Replace$(UnReturned, "\n" & vbCrLf, "\n")

End Function


' Read a string from a file, using Lfs instead of CrLfs
Public Function ReadCString(ByVal length As Long, ByVal openFile As Long) As String

    Dim A As Long
    Dim Char As String
    Dim Char2 As String
    
    
    For A = 1 To length
        Char = Input$(1, openFile)
        
        If Char = vbCr Then
            Char2 = Input$(1, openFile)
            
            ' We detected a CrLf combo
            If Char2 = vbLf Then
                Char = vbLf
            Else
                Char = Char & Char2
            End If
        End If
        
        ReadCString = ReadCString & Char
    Next A

End Function


' Pad a string with Chr$(255) characters up to a given length
Public Function Pad255(ByVal PadString As String, ByRef PadNum As Integer) As String

    If Len(PadString) < PadNum Then PadString = PadString & String$(PadNum - Len(PadString), Chr$(255))
    
    Pad255 = PadString

End Function


' Strip all &HFF characters from a string
Public Function Strip255(ByVal StripString As String) As String

    Strip255 = Trim$(Replace$(StripString, Chr$(255), Chr$(32)))

End Function


' Bounding checks for numeric textboxes to prevent some overflow errors
Public Function Bounds(ByRef txtBox As TextBox) As Long

    Dim BoundValue As String
    
    
    BoundValue = txtBox.Text
    
    ' 32-bit signed integer overflow check
    If Val(BoundValue) > ((2 ^ 31) - 1) Then BoundValue = ((2 ^ 31) - 1)
    If Val(BoundValue) < -(2 ^ 31) Then BoundValue = -(2 ^ 31)
    
    ' Update textbox if necessary
    If BoundValue <> txtBox.Text Then
        txtBox.Text = BoundValue
        txtBox.SelStart = Len(txtBox.Text)
    End If
    
    ' Return the value
    Bounds = Val(BoundValue)

End Function


' Add a file to the recents list
Public Sub AddRecentFile(ByVal filename As String)

    Dim A As Long
    Dim cCount As Long
    Dim RecentRoot As Dictionary

    
    ' Get a reference to the recent files root
    Set RecentRoot = Config.Root(True).Item("recentFiles")
    
    ' Too many on the list?
    If RecentRoot.Count > 31 Then RecentRoot.Remove RecentRoot.Keys(0)

    ' Check for any double filenames
    For A = 0 To RecentRoot.Count - 1
        If LCase$(RecentRoot.Items(A)) = LCase$(filename) Then
            ' Remove this double item (so it gets moved upwards)
            RecentRoot.Remove RecentRoot.Keys(A)
            Exit For
        End If
    Next A
    
    ' Find an unused value (bah, this sucks)
    Do
        cCount = cCount + 1
        If RecentRoot.Exists(Trim$(Str$(cCount))) = False Then Exit Do
    Loop

    ' Add the new file
    RecentRoot.Add Trim$(Str$(cCount)), filename
    
    ' Update the main form's recent file list
    frmMain.UpdateRecentList

End Sub


' Returns the dictionary of recent files (just handy to have around, but not very fast)
Public Function RecentFiles() As Dictionary

    Set RecentFiles = Config.Root(True).Item("recentFiles")

End Function


' Return a Doom\Doom2 map string from a mission and map number
Public Function MapString(ByVal Episode As Long, Map As Long) As String

    ' Doom2 (MAPxy)
    If Episode = 0 Then
        MapString = "MAP" & PadString(Trim$(Str$(Map)), "0", 2)
        Exit Function
    
    ' Doom1 (ExMy)
    Else
        MapString = "E" & Trim$(Str$(Episode)) & "M" & Trim$(Str$(Map))
        Exit Function
    
    End If

End Function


' Pad a string until it fits a given length
Public Function PadString(ByVal InText As String, ByVal padChar As String, ByVal EndLength As Long) As String

    PadString = InText
    While Len(PadString) < EndLength
        PadString = padChar & PadString
    Wend

End Function


' Return a special path
' No trailing slash is ensured
Public Function getSpecialPath(ByVal Path As specialPathEnum) As String

    Dim sPath As String
    Dim flags As Long
    

    If Path = applicationData Then
        flags = CSIDL_APPDATA
    ElseIf Path = myDocuments Then
        flags = CSIDL_MYDOCUMENTS
    ElseIf Path = userDesktop Then
        flags = CSIDL_DESKTOP
    End If
    
    sPath = String(MAX_PATH, vbNullChar)
    If Path = Temp Then
        If GetTempPath(Len(sPath), sPath) <> 0 Then getSpecialPath = TrimNull(sPath)
    Else
        If SHGetFolderPath(0, flags, 0, 0, sPath) = 0 Then getSpecialPath = TrimNull(sPath)
    End If

    If Right$(getSpecialPath, 1) = "\" Then getSpecialPath = Left$(getSpecialPath, Len(getSpecialPath) - 1)

End Function


' Initialise common controls for themed look
Public Function initCommonControls() As Boolean
   
   Dim iCCEx As tagInitCommonControlsEx
   
   On Error Resume Next
      
   
   With iCCEx
       .lngSize = LenB(iCCEx)
       .lngICC = &H200
   End With
   
   InitCommonControlsEx iCCEx
   initCommonControls = (Err.Number = 0)
   
   On Error GoTo 0
   
End Function


' String to integer conversion
Public Function CVI(ByRef Arg As String) As Integer
     
     CopyMemory ByVal VarPtr(CVI), ByVal Arg, 2
     
End Function


' String to long conversion
Public Function CVL(ByRef Arg As String) As Long

     CopyMemory ByVal VarPtr(CVL), ByVal Arg, 4
     
End Function


' Export a Doom sound to a RIFF WAVE file
Public Sub exportSoundWave(ByVal targetFile As String, ByRef data As String)

    Dim Chunk As RIFFCHUNK
    Dim waveFormat As WAVEFORMATEX
    Dim openFile As Long

    
    ' Open file for writing
    openFile = FreeFile
    Open targetFile For Binary As openFile
    
    
    ' RIFF
    Chunk.id = "RIFF"
    Chunk.Size = Len(waveFormat) + Len(data) - 8 + 20
    Put openFile, , Chunk
    Put openFile, , "WAVE"

    ' fmt
    Chunk.id = "fmt "
    Chunk.Size = Len(waveFormat)
    With waveFormat
        .wFormatTag = 1
        .nChannels = 1
        .nSamplesPerSec = CVI(Mid$(data, 3, 2))
        .wBitsPerSample = 8
        .nBlockAlign = .wBitsPerSample / 8 * .nChannels
        .nAvgBytesPerSec = .nSamplesPerSec * .nBlockAlign
    End With
    Put openFile, , Chunk
    Put openFile, , waveFormat
    
    ' data
    Chunk.id = "data"
    Chunk.Size = Len(data) - 8
    Put openFile, , Chunk
    Put openFile, , Mid$(data, 9)
    
    Close openFile
    
End Sub


' Pad the right side of a string with a character
Public Function padRight(ByVal inString As String, ByVal padChar As String, targetLength As Long) As String

    If Len(inString) < targetLength Then
        padRight = inString & String(targetLength - Len(inString), padChar)
    
    Else
        padRight = inString
        
    End If

End Function
