Attribute VB_Name = "modWADList"
' -------------------------------------------------------------------------------
'
'  WhackEd3 - A Windows Doom hacking utility
'  Copyright (C) 2003-2009  Dennis Meuwissen
'
'  This program is free software; you can redistribute it and/or modify
'  it under the terms of the GNU General Public License as published by
'  the Free Software Foundation; either version 2 of the License, or
'  (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful,
'  but WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'  GNU General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
' -------------------------------------------------------------------------------
'
'   modWADList.bas
'   Purpose: Handles WAD class lists for lump lookups.
'
Option Explicit
Option Compare Binary
Option Base 0


' Wad class list
Public WADList As New Dictionary


' Add (register) a WAD in the list
Public Function RegisterWAD(ByRef WAD As clsWad) As Boolean

    ' Exit if no WAD class
    If WAD Is Nothing Then
        RegisterWAD = False
        Exit Function
    End If
    
    ' Add the class
    WADList.Add LCase$(WAD.WadFileName), WAD
    RegisterWAD = True

End Function


' Remove (unregister) a WAD class from the list
Public Function UnRegisterWAD(ByVal filename As String) As Boolean

    ' False return by default
    UnRegisterWAD = False
    
    ' Remove it if it exists
    If WADList.Exists(LCase$(filename)) = True Then
        WADList.Remove LCase$(filename)
        UnRegisterWAD = True
    End If
    
End Function


' The cherry of this module; return a dictionary holding class and lumpnum
Public Function WADNumForName(ByVal LumpName As String) As Dictionary

    Dim A As Long
    Dim cNum As Long
    
    
    ' Return nothing by default
    Set WADNumForName = Nothing
    
    ' If no WADs in list, exit
    If WADList.Count = 0 Then Exit Function
    
    ' Step through each WAD, backwards
    For A = WADList.Count - 1 To 0 Step -1
        cNum = WADList.Items(A).NumForName(LumpName)
        
        ' If the lumpname was found, return the class and lump number in a dict.
        If cNum <> -1 Then
            Set WADNumForName = New Dictionary
            With WADNumForName
                .Add "class", WADList.Items(A)
                .Add "num", cNum
                .Add "wadIndex", A
                .Add "name", LumpName
            End With
            Exit Function
        End If
    Next A
    
End Function


' Returns a sprite lump (for sprite rotations and mirrors)
Public Function WADNumForNameSprite(ByVal SpriteName As String, ByVal FrameRot As String) As Dictionary

    Dim A As Long
    Dim cNum As Long
    
    
    ' Return nothing by default
    Set WADNumForNameSprite = Nothing
    
    ' If no WADs in list, exit
    If WADList.Count = 0 Then Exit Function
    
    ' Step through each WAD, backwards
    For A = WADList.Count - 1 To 0 Step -1
        
        cNum = WADList.Items(A).NumForSprite(SpriteName, FrameRot)
        
        ' If the lumpname was found, return the class and lump number in a dict.
        If cNum <> -1 Then
            Set WADNumForNameSprite = New Dictionary
            With WADNumForNameSprite
                .Add "class", WADList.Items(A)
                .Add "num", cNum
                .Add "name", WADList.Items(A).LumpName(cNum)
                .Add "wadIndex", A
            End With
            Exit Function
        End If
        
    Next A
    
End Function


' Clear the entire WAD list
Public Sub ClearList()

    WADList.RemoveAll

End Sub
